"""
Tagulous extensions for models which use tag fields (tagged models)

These are all applied automatically when the TAGULOUS_ENHANCE_MODELS setting
is enabled.
"""
import copy

import django
from django.core.exceptions import FieldDoesNotExist
from django.db import models, transaction

from .. import utils
from ..constants import TAGGED_ATTR_MANAGER
from .fields import (
    BaseTagField,
    SingleTagField,
    TagField,
    singletagfields_from_model,
    tagfields_from_model,
)


def _split_kwargs(model, kwargs, lookups=False, with_fields=False):
    """
    Split kwargs into fields which are safe to pass to create, and
    m2m tag fields, creating SingleTagFields as required.

    If lookups is True, TagFields with tagulous-specific lookups will also be
    matched, and the returned tag_fields will be a dict of tuples in the
    format ``(val, lookup)``

    The only tagulous-specific lookup is __exact

    For internal use only - likely to change significantly in future versions

    Returns a tuple of safe_fields, singletag_fields, tag_fields

    If with_fields is True, a fourth argument will be returned - a dict to
    look up Field objects from their names
    """
    safe_fields = {}
    singletag_fields = {}
    tag_fields = {}
    field_lookup = {}
    for field_name, val in kwargs.items():
        # Check for lookup
        if lookups and "__" in field_name:
            orig_field_name = field_name
            field_name, lookup = field_name.split("__", 1)

            # Only one known lookup
            if lookup == "exact":
                try:
                    field = model._meta.get_field(field_name)
                except FieldDoesNotExist:
                    # Unknown - pass it on untouched
                    pass
                else:
                    if isinstance(field, TagField):
                        # Store for later
                        tag_fields[field_name] = (val, lookup)
                        field_lookup[field_name] = field
                        continue

            # Irrelevant lookup - no need to take special actions
            safe_fields[orig_field_name] = val
            continue

        # No lookup
        # Try to look up the field
        try:
            field = model._meta.get_field(field_name)
        except FieldDoesNotExist:
            # Assume it's something clever and pass it through untouched
            # If it's invalid, an error will be raised later anyway
            safe_fields[field_name] = val

            # Next field
            continue

        field_lookup[field_name] = field
        # Take special measures depending on field type
        if isinstance(field, SingleTagField):
            singletag_fields[field_name] = val

        elif isinstance(field, TagField):
            # Store for later
            if lookups:
                tag_fields[field_name] = (val, None)
            else:
                tag_fields[field_name] = val

        else:
            safe_fields[field_name] = val

    if with_fields:
        return safe_fields, singletag_fields, tag_fields, field_lookup

    return safe_fields, singletag_fields, tag_fields


# ##############################################################################
# ############################################################## TaggedQuerySet
# ##############################################################################


class TaggedQuerySet(models.query.QuerySet):
    """
    A QuerySet with support for Tagulous tag fields
    """

    def _filter_or_exclude(self, negate, *args, **kwargs):
        """
        Custom lookups for tag fields
        """
        if django.VERSION >= (3, 2):
            # Arguments changed to _filter_or_exclude(self, negate, args, kwargs)
            args, kwargs = args

        # TODO: Replace with custom lookups
        safe_fields, singletag_fields, tag_fields, field_lookup = _split_kwargs(
            self.model, kwargs, lookups=True, with_fields=True
        )

        # Look up string values for SingleTagFields by name
        for field_name, val in singletag_fields.items():
            query_field_name = field_name
            if isinstance(val, str):
                query_field_name += "__name"
                if not field_lookup[field_name].tag_options.case_sensitive:
                    query_field_name += "__iexact"
            safe_fields[query_field_name] = val

        # Query as normal
        if django.VERSION >= (3, 2):
            qs = super(TaggedQuerySet, self)._filter_or_exclude(
                negate, args, safe_fields
            )
        else:
            qs = super(TaggedQuerySet, self)._filter_or_exclude(
                negate, *args, **safe_fields
            )

        # Look up TagFields by string name
        #
        # Each of these comparisons will be done with a subquery; for
        # A filter can chain, ie .filter(tags__name=..).filter(tags__name=..),
        # but exclude won't work that way; has to be done with a subquery
        for field_name, val in tag_fields.items():
            val, lookup = val
            tag_options = field_lookup[field_name].tag_options

            # Only perform custom lookup if value is a string
            if not isinstance(val, str):
                if django.VERSION >= (3, 2):
                    qs = super(TaggedQuerySet, self)._filter_or_exclude(
                        negate, [], {field_name: val}
                    )
                else:
                    qs = super(TaggedQuerySet, self)._filter_or_exclude(
                        negate, **{field_name: val}
                    )
                continue

            # Parse the tag string
            tags = utils.parse_tags(val, space_delimiter=tag_options.space_delimiter)

            # Prep the subquery
            subqs = qs
            if negate:
                subqs = self.__class__(model=self.model, using=self._db)

            # To get an exact match, filter this queryset to only include
            # items with a tag count that matches the number of specified tags
            if lookup == "exact":
                count_name = "_tagulous_count_%s" % field_name
                subqs = subqs.annotate(**{count_name: models.Count(field_name)}).filter(
                    **{count_name: len(tags)}
                )
                # Explicit order as meta ordering will be ignored
                subqs = subqs.order_by("name")

            # Prep the field name
            query_field_name = field_name + "__name"
            if not tag_options.case_sensitive:
                query_field_name += "__iexact"

            # Now chain the filters for each tag
            #
            # Have to do it this way to create new inner joins for each tag;
            # ANDing Q objects will do it all on a single inner join, which
            # will match nothing
            for tag in tags:
                subqs = subqs.filter(**{query_field_name: tag})

            # Fold subquery back into main query
            if negate:
                # Exclude on matched ID
                qs = qs.exclude(pk__in=subqs.values("pk"))
            else:
                # A filter op can just replace the main query
                qs = subqs

        return qs

    def create(self, **kwargs):
        # Create object as normal
        safe_fields, singletag_fields, tag_fields = _split_kwargs(self.model, kwargs)

        # Could convert SingleTagFields to instances with
        # field.tag_model.objects.get_or_create, but model constructor will
        # assign it through the descriptor anyway, so this is unnecessary.
        # SingleTagFields are safe
        safe_fields.update(singletag_fields)

        with transaction.atomic():
            # Create as normal
            obj = super(TaggedQuerySet, self).create(**safe_fields)

            # Add tag fields
            for field_name, val in tag_fields.items():
                setattr(obj, field_name, val)
                getattr(obj, field_name).save()

            return obj

    def get_or_create(self, defaults=None, **kwargs):
        # Get or create object as normal
        safe_fields, singletag_fields, tag_fields = _split_kwargs(self.model, kwargs)

        # As in .create, SingleTagFields are ok to create
        # Existing .get will be fine for lookup
        safe_fields.update(singletag_fields)

        # Use normal get_or_create if there are no tag fields
        if len(tag_fields) == 0:
            return super(TaggedQuerySet, self).get_or_create(defaults, **safe_fields)

        # Try to find it using get - that is able to handle tag_fields
        try:
            return self.get(**kwargs), False
        except self.model.DoesNotExist:
            return self.create(**kwargs), True

    @classmethod
    def cast_class(cls, queryset):
        """
        Changes the class of the specified queryset to a subclass of TaggedQuerySet
        and the original class, so it has all the same properties it did when it
        was first initialised, but is now a TaggedQuerySet subclass.

        The new class is given the same name as the old class, but with the prefix
        'CastTagged' to indicate the type of the object has changed, eg a normal
        QuerySet will become CastTaggedQuerySet.
        """
        # Make a subclass of TaggedQuerySet and the original class
        orig_cls = queryset.__class__
        new_cls_name = f"CastTagged{orig_cls.__name__}"
        queryset.__class__ = type(new_cls_name, (cls, orig_cls), {},)

        return queryset


# ##############################################################################
# ############################################################## TaggedManager
# ##############################################################################


class TaggedManager(models.Manager):
    """
    A manager with support for Tagulous tag fields
    """

    # The class of a tag-enabled queryset. Here so a custom TaggedManager can
    # replace it with a custom TaggedQuerySet subclass.
    tagulous_queryset = TaggedQuerySet

    def _enhance_queryset(self, qs):
        """
        Enhance an existing queryset with the class in self.tagulous_queryset
        """
        return self.tagulous_queryset.cast_class(qs)

    def get_queryset(self, *args, **kwargs):
        """
        Get the original queryset and then enhance it
        """
        qs = super(TaggedManager, self).get_queryset(*args, **kwargs)
        return self._enhance_queryset(qs)

    @classmethod
    def cast_class(cls, manager):
        """
        Changes the class of the specified manager to a subclass of TaggedManager
        and the original class, so it has all the same properties it did when it
        was first initialised, but is now a TaggedManager subclass.

        The new class is given the same name as the old class, but with the prefix
        'CastTagged' to indicate the type of the object has changed, eg a normal
        Manager will become CastTaggedManager
        """
        # Make a subclass of TaggedQuerySet and the original class
        orig_cls = manager.__class__
        new_cls_name = f"CastTagged{orig_cls.__name__}"
        manager.__class__ = type(new_cls_name, (cls, orig_cls), {})

        return manager


# ##############################################################################
# ############################################################## TaggedModel
# ##############################################################################


class TaggedModel(models.Model):
    """
    An abstract model base class with support for Tagulous tag fields
    """

    def __init__(self, *args, **kwargs):
        safe_fields, singletag_fields, tag_fields = _split_kwargs(self, kwargs)

        # Constructor has always been happy with ForeignKeys
        safe_fields.update(singletag_fields)

        # Call old init
        super(TaggedModel, self).__init__(*args, **safe_fields)

        # Add on TagField values
        for field_name, val in tag_fields.items():
            setattr(self, field_name, val)

    def __getstate__(self):
        """
        Pickle tag fields
        """
        state = super().__getstate__()

        for field in self._meta.get_fields():
            if isinstance(field, BaseTagField):
                # Remove manager from state
                attr = TAGGED_ATTR_MANAGER % field.name
                if attr in state:
                    del state[attr]

                # Add text tags
                state[field.name] = str(getattr(self, field.name))

        return state

    @classmethod
    def cast_class(cls, model):
        """
        If the model contains tag fields, change the model to subclass
        TaggedModel and enhance its managers. Called automatically on all
        models with tag fields when settings.ENHANCE_MODELS is True.

        Arguments:
            model   The model to turn into a TaggedModel subclass.
                    Will only be changed if it has tag fields.
        """
        # See if there are tag fields on this model
        tag_fields = singletagfields_from_model(model) + tagfields_from_model(model)

        # If there are no tag fields skip
        if not tag_fields:
            return

        # Ensure the model subclasses TaggedModel
        if not issubclass(model, TaggedModel):
            model.__bases__ = (TaggedModel,) + model.__bases__

        # Ensure the manager subclasses TaggedManager
        if hasattr(model, "objects") and not issubclass(
            model.objects.__class__, TaggedManager
        ):
            TaggedManager.cast_class(model.objects)

        return model

    @classmethod
    def _detag_to_serializable(cls):
        """
        Clone a fake version of this model, replacing tag fields with Field
        objects, because their to_python method will not modify arguments.

        Used by serializers to pass list values for tag fields through the
        python serializer, to be loaded back into the real tagged model when
        safe.
        """
        # Get fields on this model
        fields = cls._meta.get_fields()

        # Create a fake model
        class FakeTaggedModel(models.Model):
            def __init__(self, *args, **kwargs):
                # Django 3.2 introduced a TypeError when trying to instantiate an
                # abstract model. Set it to False to get past the check.
                if django.VERSION >= (3, 2):
                    self._meta.abstract = False

                super().__init__(*args, **kwargs)

                if django.VERSION >= (3, 2):
                    self._meta.abstract = True

            def _retag_to_original(self):
                """
                Convert this instance into an instance of the proper class it
                should have been, before _detag_to_serializable converted it.
                """
                # cls and fields from closure's scope
                data = {}
                for field in fields:
                    # ManyToOneRel and ManyToManyRel objects have no attribute
                    # contribute_to_class
                    if isinstance(field, (models.ManyToOneRel, models.ManyToManyRel)):
                        continue
                    # Find fields which are either TagFields, or not M2Ms -
                    # anything which Deserializer will have stored data for
                    elif isinstance(field, TagField) or not (
                        field.remote_field
                        and isinstance(field.remote_field, models.ManyToManyRel)
                    ):
                        # Get data from object
                        data[field.name] = getattr(self, field.name)
                return cls(**data)

            class Meta:
                abstract = True

        # Add fields to fake model
        for field in fields:
            if isinstance(field, (models.ManyToOneRel, models.ManyToManyRel)):
                # ManyToOneRel and ManyToManyRel objects have no attribute
                # contribute_to_class
                continue

            elif isinstance(field, BaseTagField):
                clone_field = models.Field(blank=field.blank, null=field.null)

            else:
                clone_field = copy.deepcopy(field)

            clone_field.contribute_to_class(FakeTaggedModel, field.name)

        FakeTaggedModel._tagulous_original_cls = cls
        return FakeTaggedModel

    class Meta:
        abstract = True



"""
Tagulous test app views
"""
from django.urls import reverse_lazy
from django.views.generic.edit import CreateView, UpdateView

from tests.tagulous_tests_app import models


def null(request):
    "Null view for reversing"
    return None


class MixedCreate(CreateView):
    model = models.SimpleMixedTest
    fields = ["name", "singletag", "tags"]
    success_url = reverse_lazy("tagulous_tests_app-null")


class MixedUpdate(UpdateView):
    model = models.SimpleMixedTest
    fields = ["name", "singletag", "tags"]
    success_url = reverse_lazy("tagulous_tests_app-null")



"""
Tagulous test: Model SingleTagField

Modules tested:
    tagulous.models.managers.SingleTagManager
    tagulous.models.descriptors.BaseTagDescriptor
    tagulous.models.descriptors.SingleTagDescriptor
    tagulous.models.fields.BaseTagField
    tagulous.models.fields.SingleTagField
"""
from django.core import exceptions
from django.db import models
from django.test import TestCase, TransactionTestCase

from tagulous import models as tag_models
from tests.lib import TagTestManager, skip_if_mysql
from tests.tagulous_tests_app import models as test_models


class ModelSingleTagFieldTest(TagTestManager, TestCase):
    """
    Test model SingleTagField
    """

    manage_models = [test_models.SingleTagFieldModel]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldModel
        self.tag_model = test_models.SingleTagFieldModel.title.tag_model
        self.tag_field = test_models.SingleTagFieldModel.title

    def test_descriptor(self):
        "Check SingleTagDescriptor is in place"
        self.assertIsInstance(self.tag_field, tag_models.SingleTagDescriptor)

    def test_tag_table(self):
        "Check the tag table exists"
        self.assertTrue(issubclass(self.tag_model, tag_models.TagModel))

    def test_empty_value(self):
        "Check the descriptor returns None for no value"
        t1 = self.create(self.test_model, name="Test")
        self.assertInstanceEqual(t1, name="Test", title=None)
        self.assertTagModel(self.tag_model, {})

    def test_tag_assign_in_constructor(self):
        "Check a tag string can be set in the constructor"
        t1 = self.test_model(name="Test", title="Mr")
        t1.save()
        self.assertEqual(t1.name, "Test")
        self.assertEqual(t1.title.name, "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_assign_by_object_in_constructor(self):
        "Check a tag object can be passed in the constructor"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model(name="Test 2", title=t1.title)
        t2.save()
        self.assertEqual(t1.title, t2.title)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertTagModel(self.tag_model, {"Mr": 2})

    def test_tag_assign_in_object_create(self):
        "Check a tag string can be passed in object.create"
        t1 = self.test_model.objects.create(name="Test", title="Mr")
        self.assertEqual(t1.name, "Test")
        self.assertEqual(t1.title.name, "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_assign_by_object_in_object_create(self):
        "Check a tag object can be passed in object.create"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.create(name="Test 2", title=t1.title)
        self.assertEqual(t1.title, t2.title)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertTagModel(self.tag_model, {"Mr": 2})

    def test_tag_assign_in_object_get_or_create_true(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does not exist
        """
        t1, state = self.test_model.objects.get_or_create(name="Test", title="Mr")
        self.assertEqual(state, True)
        self.assertEqual(t1.name, "Test")
        self.assertEqual(t1.title.name, "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_tag_assign_in_object_get_or_create_false(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does exist
        """
        t1 = self.test_model.objects.create(name="Test", title="Mr")
        t2, state = self.test_model.objects.get_or_create(name="Test", title="Mr")
        self.assertEqual(state, False)
        self.assertEqual(t1.pk, t2.pk)
        self.assertEqual(t1.name, t2.name)
        self.assertEqual(t2.name, "Test")
        self.assertEqual(t2.title.name, "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_tag_assign_by_string(self):
        "Check a tag string can be assigned to descriptor and returned"
        t1 = self.create(self.test_model, name="Test")
        t1.title = "Mr"

        # Returned before save
        self.assertEqual(t1.title.__class__, self.tag_model)
        self.assertEqual(t1.title.name, "Mr")
        self.assertEqual(str(t1.title), "Mr")
        self.assertTagModel(self.tag_model, {})

        # Returned after save
        t1.save()
        self.assertEqual(t1.title.__class__, self.tag_model)
        self.assertEqual(t1.title.name, "Mr")
        self.assertEqual(str(t1.title), "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_assign_by_object(self):
        """
        Check a tag object can be assigned to a SingleTagfield, and that its
        tag count is incremented
        """
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model(name="Test 2")
        t2.title = t1.title
        t2.save()
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t2.name, "Test 2")
        self.assertEqual(str(t1.title), "Mr")
        self.assertEqual(str(t2.title), "Mr")
        self.assertEqual(t1.title, t2.title)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertTagModel(self.tag_model, {"Mr": 2})

    def test_assign_string_empty(self):
        "Check an empty string can clear a SingleTagField"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        self.assertInstanceEqual(t1, name="Test 1", title="Mr")
        t1.title = ""
        t1.save()
        self.assertInstanceEqual(t1, name="Test 1", title=None)
        self.assertTagModel(self.tag_model, {})

    def test_assign_none(self):
        "Check assigning None can clear a SingleTagField"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        self.assertInstanceEqual(t1, name="Test 1", title="Mr")
        t1.title = None
        t1.save()
        self.assertInstanceEqual(t1, name="Test 1", title=None)
        self.assertTagModel(self.tag_model, {})

    def test_assign_string_quotes(self):
        "Check that a tag can contain quotes"
        # Check it saves ok
        t1 = self.test_model.objects.create(name="Test 1", title='"One", Two"')
        self.assertInstanceEqual(t1, name="Test 1", title='"One", Two"')

        # Check it loads ok
        t2 = self.test_model.objects.get(pk=t1.pk)
        self.assertInstanceEqual(t2, name="Test 1", title='"One", Two"')

    def test_tag_assign_same(self):
        "Check that setting the same tag doesn't make a change"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        tag = t1.title
        self.assertEqual(tag, "Mr")

        # Access the manager via the cache to watch changed state
        self.assertIsInstance(t1._title_tagulous, tag_models.SingleTagManager)
        self.assertFalse(t1._title_tagulous.changed, "tag has changed")
        t1.title = "Mr"
        self.assertFalse(t1._title_tagulous.changed, "tag has changed")

    def test_change_decreases_count(self):
        "Check a tag string changes the count"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.create(name="Test 2", title=t1.title)
        self.assertTagModel(self.tag_model, {"Mr": 2})
        t2.title = "Mrs"
        t2.save()
        self.assertTagModel(self.tag_model, {"Mr": 1, "Mrs": 1})

    def test_delete_decreases_count(self):
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.create(name="Test 2", title=t1.title)
        self.assertTagModel(self.tag_model, {"Mr": 2})
        t2.delete()
        self.assertInstanceEqual(t1, name="Test 1", title="Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_count_zero_deletes_tag(self):
        "Check a count of 0 deletes an unprotected tag"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})
        t1.title = "Mrs"
        t1.save()
        self.assertTagModel(self.tag_model, {"Mrs": 1})

    def test_delete_decreases_correct(self):
        """
        Check that the actual tag in the database is decreased, not the one in
        the instance at time of deletion
        """
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.create(name="Test 2", title="Mrs")
        self.assertEqual(str(t1.title.name), "Mr")
        self.assertEqual(str(t2.title.name), "Mrs")
        self.assertTagModel(self.tag_model, {"Mr": 1, "Mrs": 1})

        # Now change the title and delete without saving
        t1.title = "Mrs"
        self.assertEqual(str(t1.title.name), "Mrs")
        self.assertTagModel(self.tag_model, {"Mr": 1, "Mrs": 1})
        t1.delete()

        # Check the original tag 'Mr' was decremented (and deleted)
        self.assertTagModel(self.tag_model, {"Mrs": 1})

        # But check that tagulous still thinks the tag is 'Mrs'
        self.assertEqual(str(t1.title.name), "Mrs")

    def test_save_deleted_instance(self):
        """
        Check that a deleted tag in memory can be re-saved when the instance it
        is set on is deleted
        """
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})
        t1.delete()
        self.assertTagModel(self.tag_model, {})
        t1.save()
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_save_deleted_tag(self):
        """
        Check that a delete tag in memory can be read and re-saved when it is
        deleted without the instance knowing about it
        """
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

        self.tag_model.objects.all().delete()
        self.assertTagModel(self.tag_model, {})

        # Check it's still usable
        self.assertIsInstance(t1.title, tag_models.BaseTagModel)
        self.assertEqual(t1.title, "Mr")

        # Check it can be re-saved
        t1.save()
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_multiple_unsaved(self):
        "Check that there's no leak between unsaved objects"
        t1 = self.test_model(name="Test 1", title="Mr")
        t2 = self.test_model(name="Test 1", title="Mrs")
        self.assertTagModel(self.tag_model, {})
        self.assertEqual(str(t1.title), "Mr")
        self.assertEqual(str(t2.title), "Mrs")
        t1.save()
        self.assertTagModel(self.tag_model, {"Mr": 1})
        t2.save()
        self.assertTagModel(self.tag_model, {"Mr": 1, "Mrs": 1})
        self.assertEqual(str(t1.title), "Mr")
        self.assertEqual(str(t2.title), "Mrs")

    def test_load_instance(self):
        "Check that SingleTagField is loaded correctly"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.get(pk=t1.pk)
        self.assertIsInstance(t2.title, tag_models.BaseTagModel)
        self.assertEqual(t1.title, t2.title)
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_descriptor_equal(self):
        "Check that descriptors evaluate to equal"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.create(name="Test 2", title="Mr")
        self.assertIsInstance(t1.title, self.tag_model)
        self.assertEqual(t1.title.pk, t2.title.pk)
        self.assertEqual(t1.title, t2.title)

    def test_descriptor_not_equal(self):
        "Check that descriptors evaluate to equal"
        t1 = self.test_model.objects.create(name="Test 1", title="Mr")
        t2 = self.test_model.objects.create(name="Test 2", title="Mrs")
        self.assertIsInstance(t1.title, self.tag_model)
        self.assertNotEqual(t1.title, t2.title)

    def test_cascade_delete(self):
        "Check that deleting a tag deletes its related tagged items (by default)"
        model = self.test_model
        t1 = model.objects.create(name="Test 1", title="Mr")
        t2 = model.objects.create(name="Test 2", title="Mr")
        t3 = model.objects.create(name="Test 3", title="Mrs")
        self.assertTagModel(self.tag_model, {"Mr": 2, "Mrs": 1})
        self.assertSequenceEqual(model.objects.all(), [t1, t2, t3])

        # Delete
        self.tag_model.objects.get(name="Mr").delete()
        self.assertTagModel(self.tag_model, {"Mrs": 1})
        self.assertSequenceEqual(model.objects.all(), [t3])


# ##############################################################################
# ######  Test it works with concrete inheritance
# ##############################################################################


class ModelSingleTagFieldConcreteInheritanceTest(ModelSingleTagFieldTest):
    manage_models = [test_models.SingleTagFieldConcreteInheritanceModel]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldConcreteInheritanceModel
        self.tag_model = (
            test_models.SingleTagFieldConcreteInheritanceModel.title.tag_model
        )
        self.tag_field = test_models.SingleTagFieldConcreteInheritanceModel.title

    def test_save_deleted_instance(self):
        """
        Disable this test as it doesn't make sense in the context of a concrete
        subclass; when the subclass is deleted, it will delete the base model instance,
        meaning when we go to save it again we'll get the error:

            save() prohibited to prevent data loss due to unsaved related object ..._ptr

        where ..._ptr is a OneToOneField to the base model. This is expected behaviour,
        and out of scope for this package to address.
        """


# ##############################################################################
# ######  Test it works with abstract inheritance
# ##############################################################################


class ModelSingleTagFieldAbstractInheritanceTest(
    ModelSingleTagFieldConcreteInheritanceTest
):
    manage_models = [test_models.SingleTagFieldAbstractInheritanceModel]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldAbstractInheritanceModel
        self.tag_model = (
            test_models.SingleTagFieldAbstractInheritanceModel.title.tag_model
        )
        self.tag_field = test_models.SingleTagFieldAbstractInheritanceModel.title

    def test_second_inheritance__tag_model_not_shared(self):
        second_test_model = test_models.SingleTagFieldAbstractInheritanceSecondModel
        second_tag_model = second_test_model.title.tag_model
        self.assertNotEqual(self.tag_model, second_tag_model)

    def test_second_inheritance__writes_update_shared_model(self):
        second_test_model = test_models.SingleTagFieldAbstractInheritanceSecondModel
        second_tag_model = second_test_model.title.tag_model

        t1 = self.test_model(name="Test 1", title="Mr")
        t1.save()

        t2 = second_test_model(name="Test 2", title="Mrs")
        t2.save()

        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.title.name, "Mr")
        self.assertEqual(t2.name, "Test 2")
        self.assertEqual(t2.title.name, "Mrs")
        self.assertTagModel(self.tag_model, {"Mr": 1})
        self.assertTagModel(second_tag_model, {"Mrs": 1})


# ##############################################################################
# ######  Test invalid fields
# ##############################################################################


class ModelSingleTagFieldInvalidTest(TagTestManager, TransactionTestCase):
    """
    Test invalid model SingleTagField

    Use a TransactionTestCase so the apps will be reset
    """

    manage_models = [test_models.SingleTagFieldModel]

    def test_invalid_to_model(self):
        "Check that the to model has to be a TagModel subclass"
        with self.assertRaises(ValueError) as cm:

            class FailModel_invalid_to(models.Model):
                to_model = tag_models.SingleTagField(test_models.SingleTagFieldModel)

        self.assertEqual(str(cm.exception), "Tag model must be a subclass of TagModel")

    def test_forbidden_to_field(self):
        "Check that to_field argument raises exception"
        with self.assertRaises(ValueError) as cm:

            class FailModel_forbidden_to(models.Model):
                to_field = tag_models.SingleTagField(to_field="fail")

        self.assertEqual(
            str(cm.exception), "Invalid argument 'to_field' for SingleTagField",
        )

    def test_forbidden_rel_class(self):
        "Check that rel_class argument raises exception"
        with self.assertRaises(ValueError) as cm:

            class FailModel_forbidden_rel(models.Model):
                rel_class = tag_models.SingleTagField(rel_class="fail")

        self.assertEqual(
            str(cm.exception), "Invalid argument 'rel_class' for SingleTagField",
        )

    def test_forbidden_max_count(self):
        "Check that max_count argument raises exception"
        with self.assertRaises(ValueError) as cm:

            class FailModel_forbidden_max_count(models.Model):
                max_count = tag_models.SingleTagField(max_count="fail")

        self.assertEqual(
            str(cm.exception), "Invalid argument 'max_count' for SingleTagField",
        )

    def test_value_from_object_none(self):
        "Check that value_from_object returns empty string instead of None"
        # Called by forms, but part of model, so test it here
        t1 = test_models.SingleTagFieldModel.objects.create(name="Test 1")
        field = test_models.SingleTagFieldModel._meta.get_field("title")
        self.assertEqual(field.value_from_object(t1), "")


# ##############################################################################
# ######  Test model field blank=True
# ##############################################################################


class ModelSingleTagFieldOptionalTest(TagTestManager, TestCase):
    """
    Test optional model SingleTagField
    """

    manage_models = [test_models.SingleTagFieldOptionalModel]

    def test_optional_save_missing(self):
        "Check an optional SingleTagField isn't required for save"
        # If it fails, it will return an exception
        t1 = test_models.SingleTagFieldOptionalModel(name="Test 1")
        t1.save()
        self.assertNotEqual(t1.pk, None)

    def test_optional_create_missing(self):
        "Check an optional SingleTagField isn't required for object.create"
        # If it fails, it will return an exception
        t1 = test_models.SingleTagFieldOptionalModel.objects.create(name="Test 1")
        self.assertNotEqual(t1.pk, None)


# ##############################################################################
# ###### Test model field blank=False
# ##############################################################################


class ModelSingleTagFieldRequiredTest(TagTestManager, TestCase):
    """
    Test required model SingleTagField
    """

    manage_models = [test_models.SingleTagFieldRequiredModel]

    def test_required_save_raises(self):
        "Check a required SingleTagField raises an exception when saved"
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.SingleTagFieldRequiredModel(name="Test")
            t1.save()
        self.assertEqual(cm.exception.messages[0], "This field cannot be null.")

    def test_required_create_raises(self):
        "Check a required SingleTagField raises an exception in object.create"
        with self.assertRaises(exceptions.ValidationError) as cm:
            test_models.SingleTagFieldRequiredModel.objects.create(name="Test")
        self.assertEqual(cm.exception.messages[0], "This field cannot be null.")


# ##############################################################################
# ###### Test multiple SingleTagFields on a model
# ##############################################################################


class ModelSingleTagFieldMultipleTest(TagTestManager, TestCase):
    """
    Test multiple tag fields on a model
    """

    manage_models = [test_models.SingleTagFieldMultipleModel]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldMultipleModel
        self.tag_field_1 = test_models.SingleTagFieldMultipleModel.tag1
        self.tag_field_2 = test_models.SingleTagFieldMultipleModel.tag2
        self.tag_field_3 = test_models.SingleTagFieldMultipleModel.tag3

    def test_tag_models(self):
        self.assertTrue(issubclass(self.tag_field_1.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_2.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_3.tag_model, tag_models.TagModel))

    def test_separate_models(self):
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_2.tag_model)
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_3.tag_model)
        self.assertNotEqual(self.tag_field_2.tag_model, self.tag_field_3.tag_model)

    def test_model_names(self):
        self.assertEqual(
            self.tag_field_1.tag_model.__name__,
            "Tagulous_SingleTagFieldMultipleModel_tag1",
        )
        self.assertEqual(
            self.tag_field_2.tag_model.__name__,
            "Tagulous_SingleTagFieldMultipleModel_tag2",
        )
        self.assertEqual(
            self.tag_field_3.tag_model.__name__,
            "Tagulous_SingleTagFieldMultipleModel_tag3",
        )

    def test_set_and_get(self):
        "Test multiple fields can be set and retrieved independently"
        t1 = self.create(
            self.test_model, name="Test 1", tag1="Mr", tag2="blue", tag3="adam"
        )
        t2 = self.create(
            self.test_model, name="Test 2", tag1="Mrs", tag2="green", tag3="brian"
        )
        t3 = self.create(
            self.test_model, name="Test 3", tag1="Ms", tag2="red", tag3="chris"
        )

        self.assertTagModel(self.tag_field_1, {"Mr": 1, "Mrs": 1, "Ms": 1})
        self.assertTagModel(self.tag_field_2, {"blue": 1, "green": 1, "red": 1})
        self.assertTagModel(self.tag_field_3, {"adam": 1, "brian": 1, "chris": 1})

        self.assertInstanceEqual(t1, name="Test 1", tag1="Mr", tag2="blue", tag3="adam")
        self.assertInstanceEqual(
            t2, name="Test 2", tag1="Mrs", tag2="green", tag3="brian"
        )
        self.assertInstanceEqual(t3, name="Test 3", tag1="Ms", tag2="red", tag3="chris")


# ##############################################################################
# ###### Test SingleTagField with string references to tag model
# ##############################################################################


class ModelSingleTagFieldStringTest(TagTestManager, TransactionTestCase):
    """
    Test SingleTagField which refers to its tag model with a string
    """

    manage_models = [test_models.MixedStringTo]

    def setUpExtra(self):
        self.test_model = test_models.MixedStringTo
        self.tag_field = self.test_model.singletag
        self.tag_model = test_models.MixedStringTagModel

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        self.assertEqual(self.tag_field.field.remote_field.model, self.tag_model)
        self.assertEqual(self.tag_field.tag_model, self.tag_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(self.tag_field.tag_options, self.tag_model.tag_options)

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.tag_model, {})
        t1 = self.test_model.objects.create(name="Test 1", singletag="Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})
        self.assertInstanceEqual(t1, name="Test 1", singletag="Mr")


class ModelSingleTagFieldSelfTest(TagTestManager, TransactionTestCase):
    """
    Test SingleTagField which refers to itself
    """

    manage_models = [test_models.MixedSelfTo]

    def setUpExtra(self):
        self.test_model = test_models.MixedSelfTo
        self.tag_field = self.test_model.alternate

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        self.assertEqual(self.tag_field.field.remote_field.model, self.test_model)
        self.assertEqual(self.tag_field.tag_model, self.test_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(self.tag_field.tag_options, self.test_model.tag_options)

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.test_model, {})
        t1 = self.test_model.objects.create(name="Test 1", alternate="mr")
        self.assertTagModel(self.test_model, {"mr": 1, "Test 1": 0})
        self.assertInstanceEqual(t1, name="Test 1", alternate="mr")


# ##############################################################################
# ###### Test SingleTagField options
# ##############################################################################


class ModelSingleTagFieldOptionsTest(TagTestManager, TestCase):
    """
    Test single tag field options
    """

    manage_models = [test_models.SingleTagFieldOptionsModel]

    def setUpExtra(self):
        self.test_model = test_models.SingleTagFieldOptionsModel

    def test_initial_string(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(
            self.test_model.initial_string, {"Mr": 0, "Mrs": 0, "Ms": 0}
        )

    def test_initial_list(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(self.test_model.initial_list, {"Mr": 0, "Mrs": 0, "Ms": 0})

    def test_protect_initial_true(self):
        self.assertTagModel(self.test_model.protect_initial_true, {"Mr": 0})

        t1 = self.create(self.test_model, name="Test 1", protect_initial_true="Mr")
        self.assertTagModel(self.test_model.protect_initial_true, {"Mr": 1})

        t1.protect_initial_true = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_true, {"Mr": 0})

    def test_protect_initial_false(self):
        self.assertTagModel(self.test_model.protect_initial_true, {"Mr": 0})

        t1 = self.create(self.test_model, name="Test 1", protect_initial_false="Mr")
        self.assertTagModel(self.test_model.protect_initial_false, {"Mr": 1})

        t1.protect_initial_false = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_false, {})

    def test_protect_all_true(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_true="Mr")
        self.assertTagModel(self.test_model.protect_all_true, {"Mr": 1})

        t1.protect_all_true = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_all_true, {"Mr": 0})

    def test_protect_all_false(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_false="Mr")
        self.assertTagModel(self.test_model.protect_all_false, {"Mr": 1})

        t1.protect_all_false = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_all_false, {})

    @skip_if_mysql
    def test_case_sensitive_true(self):
        self.assertTagModel(self.test_model.case_sensitive_true, {"Mr": 0})
        self.create(self.test_model, name="Test 1", case_sensitive_true="mr")
        self.assertTagModel(self.test_model.case_sensitive_true, {"Mr": 0, "mr": 1})

    @skip_if_mysql
    def test_case_sensitive_false(self):
        self.assertTagModel(self.test_model.case_sensitive_false, {"Mr": 0})
        self.create(self.test_model, name="Test 1", case_sensitive_false="mr")
        self.assertTagModel(self.test_model.case_sensitive_false, {"Mr": 1})

    def test_force_lowercase_true(self):
        self.create(self.test_model, name="Test 1", force_lowercase_true="Mr")
        self.assertTagModel(self.test_model.force_lowercase_true, {"mr": 1})

    def test_force_lowercase_false(self):
        self.create(self.test_model, name="Test 1", force_lowercase_false="Mr")
        self.assertTagModel(self.test_model.force_lowercase_false, {"Mr": 1})



"""
Tagulous test: View-related functionality

Modules tested:
    tagulous.views
"""
import json

from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.urls import reverse

from tagulous import models as tag_models
from tests.lib import TagTestManager, skip_if_mysql
from tests.tagulous_tests_app import models as test_models


client = Client()


# ##############################################################################
# ###### Test tag forms with CBVs
# ##############################################################################


class TagFormCBVTest(TagTestManager, TestCase):
    "Test CBVs which use tagged forms"
    manage_models = [test_models.SimpleMixedTest]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.model = test_models.SimpleMixedTest
        self.singletag_model = self.model.singletag.tag_model
        self.tags_model = self.model.tags.tag_model

    def test_cbv_create(self):
        "Test CBV create"
        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.tags_model, {})

        # Create
        response = client.post(
            reverse("tagulous_tests_app-MixedCreate"),
            {"name": "Test 1", "singletag": "Mr", "tags": "blue, red"},
        )
        self.assertEqual(response.status_code, 302)

        # Get
        t1 = self.model.objects.get(name="Test 1")
        self.assertIsInstance(t1.singletag, tag_models.BaseTagModel)
        self.assertEqual(t1.singletag.name, "Mr")
        self.assertEqual(t1.tags.count(), 2)
        self.assertEqual(t1.tags.all()[0], "blue")
        self.assertEqual(t1.tags.all()[1], "red")
        self.assertTagModel(self.singletag_model, {"Mr": 1})
        self.assertTagModel(self.tags_model, {"blue": 1, "red": 1})

    def test_cbv_update(self):
        "Test CBV update"
        # Set up object
        t1 = self.model.objects.create(name="Test 1", singletag="Mr", tags="blue, red")
        self.assertTagModel(self.singletag_model, {"Mr": 1})
        self.assertTagModel(self.tags_model, {"blue": 1, "red": 1})

        # Update it
        response = client.post(
            reverse("tagulous_tests_app-MixedUpdate", kwargs={"pk": t1.pk}),
            {"name": "Test 2", "singletag": "Mrs", "tags": "blue, green"},
        )
        self.assertEqual(response.status_code, 302)

        # Check
        t2 = self.model.objects.get(pk=t1.pk)
        self.assertIsInstance(t1.singletag, tag_models.BaseTagModel)
        self.assertEqual(t2.name, "Test 2")
        self.assertEqual(t2.singletag.name, "Mrs")
        self.assertEqual(t2.tags.count(), 2)
        self.assertEqual(t2.tags.all()[0], "blue")
        self.assertEqual(t2.tags.all()[1], "green")
        self.assertTagModel(self.singletag_model, {"Mrs": 1})
        self.assertTagModel(self.tags_model, {"blue": 1, "green": 1})


# ##############################################################################
# ###### Test autocomplete view
# ##############################################################################


def get_response_content(response):
    return response.content.decode("utf-8")


class AutocompleteViewTest(TagTestManager, TestCase):
    "Test autocomplete view"
    manage_models = [test_models.TagFieldOptionsModel]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.test_model = test_models.TagFieldOptionsModel

    def test_unlimited(self):
        "Test unlimited autocomplete view"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(reverse("tagulous_tests_app-unlimited"))
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 100)
        for i in range(100):
            self.assertEqual(data["results"][i], "tag%02d" % i)
        self.assertEqual(data["more"], False)

    def test_unlimited_query(self):
        "Test unlimited autocomplete view with query"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(reverse("tagulous_tests_app-unlimited"), {"q": "tag0"})
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 10)
        for i in range(10):
            self.assertEqual(data["results"][i], "tag%02d" % i)
        self.assertEqual(data["more"], False)

    def test_unlimited_query__contains(self):
        "Test unlimited autocomplete view with query and contains"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Force settings
        tag_model.tag_options.autocomplete_view_fulltext = True

        # Get them from view
        response = client.get(reverse("tagulous_tests_app-unlimited"), {"q": "ag0"})
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 10)
        for i in range(10):
            self.assertEqual(data["results"][i], "tag%02d" % i)
        self.assertEqual(data["more"], False)

    def test_limited(self):
        "Test limited autocomplete view"
        # Add some tags
        tag_model = self.test_model.autocomplete_limit.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        page_length = tag_model.tag_options.autocomplete_limit
        self.assertEqual(page_length, 3)

        # Get page 1
        response = client.get(reverse("tagulous_tests_app-limited"))
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), page_length)
        self.assertEqual(data["results"][0], "tag00")
        self.assertEqual(data["results"][1], "tag01")
        self.assertEqual(data["results"][2], "tag02")
        self.assertEqual(data["more"], True)

        # Get page 4: starting 10th tag, tag09 to tag11
        page = 4
        response = client.get(reverse("tagulous_tests_app-limited"), {"p": page})
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), page_length)
        self.assertEqual(data["results"][0], "tag09")
        self.assertEqual(data["results"][1], "tag10")
        self.assertEqual(data["results"][2], "tag11")
        self.assertEqual(data["more"], True)

        # Get last page, 34: starting 100th tag, tag99
        page = 34
        response = client.get(reverse("tagulous_tests_app-limited"), {"p": page})
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0], "tag99")
        self.assertEqual(data["more"], False)

    def test_limited_query(self):
        "Test limited autocomplete view with query"
        # Add some tags
        tag_model = self.test_model.autocomplete_limit.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        page_length = tag_model.tag_options.autocomplete_limit
        self.assertEqual(page_length, 3)

        # Get page 1
        response = client.get(reverse("tagulous_tests_app-limited"), {"q": "tag1"})
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), page_length)
        self.assertEqual(data["results"][0], "tag10")
        self.assertEqual(data["results"][1], "tag11")
        self.assertEqual(data["results"][2], "tag12")
        self.assertEqual(data["more"], True)

        # Get last page, 4: starting 10th tag, tag19
        page = 4
        response = client.get(
            reverse("tagulous_tests_app-limited"), {"q": "tag1", "p": page}
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0], "tag19")
        self.assertEqual(data["more"], False)

    def test_login(self):
        "Test autocomplete_login view"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        User.objects.create_user("test", "test@example.com", "password")
        client.login(username="test", password="password")
        response = client.get(reverse("tagulous_tests_app-login"))
        client.logout()
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 100)
        for i in range(100):
            self.assertEqual(data["results"][i], "tag%02d" % i)
        self.assertEqual(data["more"], False)

    def test_queryset(self):
        "Test autocomplete view on a tag model queryset"
        # Add some tags
        tag_model = self.test_model.autocomplete_view.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(reverse("tagulous_tests_app-queryset"))
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 10)
        for i in range(10):
            self.assertEqual(data["results"][i], "tag2%d" % i)
        self.assertEqual(data["more"], False)

    def test_force_lowercase_true(self):
        "Test autocomplete view on a tag model with force_lowercase=True"
        # Add some tags
        tag_model = self.test_model.force_lowercase_true.tag_model
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse("tagulous_tests_app-force_lowercase_true"), {"q": "Tag1"}
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 10)
        for i in range(10):
            self.assertEqual(data["results"][i], "tag1%d" % i)
        self.assertEqual(data["more"], False)

    @skip_if_mysql
    def test_case_sensitive_false(self):
        "Test autocomplete view on a tag model with case_sensitive=False"
        # Add some tags
        tag_model = self.test_model.case_sensitive_false.tag_model
        tag_model.objects.all().delete()  # clear out initial
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse("tagulous_tests_app-case_sensitive_false"), {"q": "Tag1"}
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))
        self.assertEqual(len(data["results"]), 10)
        for i in range(10):
            self.assertEqual(data["results"][i], "tag1%d" % i)
        self.assertEqual(data["more"], False)

    def test_case_sensitive_true(self):
        "Test autocomplete view on a tag model with case_sensitive=True"
        # Add some tags
        tag_model = self.test_model.case_sensitive_true.tag_model
        tag_model.objects.all().delete()  # clear out initial
        for i in range(100):
            tag_model.objects.create(name="tag%02d" % i)
        self.assertEqual(tag_model.objects.count(), 100)

        # Get them from view
        response = client.get(
            reverse("tagulous_tests_app-case_sensitive_true"), {"q": "Tag1"}
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(get_response_content(response))

        from django.db import connection

        if connection.vendor == "sqlite":
            # Sqlite doesn't support case insensitive searches - expect to
            # get back 10 matches
            self.assertEqual(len(data["results"]), 10)
            for i in range(10):
                self.assertEqual(data["results"][i], "tag1%d" % i)
        else:
            self.assertEqual(len(data["results"]), 0)
            self.assertEqual(data["more"], False)



import os
import re
import sys
import unittest
from io import StringIO

import django
from django.db import connection, models
from django.test import testcases

from tagulous import models as tag_models


# Detect test environment
# This is used when creating files (migrations and fixtures) to ensure that
# tests can be run concurrently without interfering with each other.
# Also calculated in setup.py
testenv = re.sub(
    r"[^a-zA-Z0-9]",
    "_",
    os.environ.get("TOXENV", "_".join(str(v) for v in django.VERSION)),
)


class TagTestManager(object):
    """
    Test mixin to help test tag models
    """

    # We have some very long string comparisons (eg form field renders,
    # migrations), so set the maxDiff to 10k
    maxDiff = 1024 * 20

    # This class can manage models
    manage_models = None

    # Show normal fail messages in addition to any custom
    longMessage = True

    def setUp(self):
        """
        Ensure initial data is in the tag models
        """
        if self.manage_models is not None:
            for model in self.manage_models:
                tag_models.initial.model_initialise_tags(model)
                tag_models.initial.model_initialise_tags(model)

        if hasattr(self, "setUpExtra"):
            self.setUpExtra()

    def tearDown(self):
        """
        Common tear down operations
        """
        # Only here for consistency
        if hasattr(self, "tearDownExtra"):
            self.tearDownExtra()

    def create(self, model, **kwargs):
        # This could be replaced with the Tagulous create() method, but do it
        # manually to avoid complicating tests
        normal = {}
        tagfield = {}
        mmfield = {}
        for field_name, val in kwargs.items():
            if isinstance(model._meta.get_field(field_name), tag_models.TagField):
                tagfield[field_name] = val
            elif isinstance(model._meta.get_field(field_name), models.ManyToManyField):
                mmfield[field_name] = val
            else:
                normal[field_name] = val

        # Create as normal
        item = model.objects.create(**normal)

        # Add tagfields (may not be using enhanced queryset)
        for field_name, val in tagfield.items():
            setattr(item, field_name, val)
            getattr(item, field_name).save()

        # Add mm fields
        for field_name, val in mmfield.items():
            field = getattr(item, field_name)
            for obj in val:
                field.add(obj)

        return item

    def assertRegex(self, text, expected_regex, msg="Regex didn't match"):
        # Implement python 3.2's assertRegex for all
        if isinstance(text, str):
            expected_regex = re.compile(expected_regex)
        self.assertTrue(expected_regex.search(text), msg=msg)

    def assertInstanceEqual(self, instance, **kwargs):
        # First, reload instance
        instance = instance.__class__.objects.get(pk=instance.pk)

        # Check values
        for field_name, val in kwargs.items():
            try:
                if isinstance(
                    instance.__class__._meta.get_field(field_name),
                    (tag_models.SingleTagField, tag_models.TagField),
                ) and isinstance(val, str):
                    self.assertEqual(str(getattr(instance, field_name)), val)
                elif isinstance(
                    instance.__class__._meta.get_field(field_name),
                    models.ManyToManyField,
                ):
                    mm_objs = list(getattr(instance, field_name).all())
                    self.assertEqual(len(val), len(mm_objs))
                    for obj in val:
                        self.assertTrue(obj in mm_objs)
                else:
                    self.assertEqual(getattr(instance, field_name), val)
            except AssertionError as e:
                self.fail("Instances not equal for field %s: %s" % (field_name, e))

    def assertTagModel(self, model, tag_counts):
        """
        Assert the tag model matches the specified tag counts
        """
        if isinstance(
            model, (tag_models.SingleTagDescriptor, tag_models.TagDescriptor)
        ):
            model = model.tag_model

        if len(tag_counts) != model.objects.count():
            self.fail(
                "Incorrect number of tags in '%s'; expected %d, got %d"
                % (model, len(tag_counts), model.objects.count())
            )

        for tag_name, count in tag_counts.items():
            try:
                tag = model.objects.get(name=tag_name)
            except model.DoesNotExist:
                self.fail("Tag model missing expected tag '%s'" % tag_name)
            if tag.count != count:
                self.fail(
                    "Tag count for '%s' incorrect; expected %d, got %d"
                    % (tag_name, count, tag.count)
                )

    def _extract_json(self, dom, path=""):
        "Recursively break out json from a django.utils.html_parser dom object"
        jsons = {}
        if isinstance(dom, str):
            return dom, {}
        el_name = path + "." + dom.name

        # Extract json from this element
        for i, attr in enumerate(dom.attributes):
            attr_name, attr_val = attr
            if not attr_name.startswith("data-"):
                continue

            if (attr_val[0] == "{" and attr_val[-1] == "}") or (
                attr_val[0] == "[" and attr_val[-1] == "]"
            ):
                if el_name not in jsons:
                    jsons[el_name] = {}
                jsons[el_name][attr_name] = attr_val
                dom.attributes[i] = (attr_name, "/*json*/")

        # Look at children
        for child in dom.children:
            new_child, new_jsons = self._extract_json(child, el_name)
            jsons.update(new_jsons)
        return dom, jsons

    def assertHTMLEqual(self, html1, html2, msg=None):
        """
        Clone of django's method, but with support for JSON in data- tag
        attributes and backwards-compatible support for HTML 5 "required"
        """
        # Add backwards compatibility for HTML 5 "required"
        REQUIRED = "{{required}}"
        required = " required "
        if REQUIRED in html1:
            html1 = html1.replace(REQUIRED, required)
        if REQUIRED in html2:
            html2 = html2.replace(REQUIRED, required)

        # Parse the HTML into a dom
        err1 = "First argument is not a valid HTML element:"
        err2 = "Second argument is not a valid HTML element:"
        dom1 = testcases.assert_and_parse_html(self, html1, msg, err1)
        dom2 = testcases.assert_and_parse_html(self, html2, msg, err2)

        # Check it actually managed to parse something
        self.assertNotEqual(dom1.name, None, msg="%s\n%s" % (err1, html1))
        self.assertNotEqual(dom2.name, None, msg="%s\n%s" % (err2, html2))

        # Walk the trees and pull out json
        dom1, json1 = self._extract_json(dom1)
        dom2, json2 = self._extract_json(dom2)

        # Convert dom back to string, call super to test doms
        # Yes it's inefficient, but it's tests and saves me from forking it
        super(TagTestManager, self).assertHTMLEqual(str(dom1), str(dom2))

        # Test jsons
        # Assert we've found the same elements
        self.assertSequenceEqual(json1.keys(), json2.keys())
        for dom_path in json1.keys():
            # Assert the element has the same attributes
            self.assertSequenceEqual(
                json1[dom_path].keys(),
                json2[dom_path].keys(),
                msg="%s has attributes do not match: %r != %r"
                % (dom_path, json1[dom_path].keys(), json2[dom_path].keys()),
            )
            for attr_name in json1[dom_path].keys():
                self.assertJSONEqual(
                    json1[dom_path][attr_name], json2[dom_path][attr_name]
                )

    def debugTagModel(self, model):
        """
        Print tag model tags and their counts, to help debug failed tests
        """
        print("-=-=-=-=-=-")
        if isinstance(
            model, (tag_models.SingleTagDescriptor, tag_models.TagDescriptor)
        ):
            model = model.tag_model
        print("Tag model: %s" % model)
        for tag in model.objects.all():
            print("%s: %d" % (tag.name, tag.count))
        print("-=-=-=-=-=-")

    def prettyPrint(self, data):
        "Pretty print data, to help debug failed tests"
        import pprint

        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(data)


# Based on http://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call
class Capturing(list):
    "Capture stdout and stderr to a string"

    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = sys.stderr = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        # Ensure output is unicode
        self.extend(str(line) for line in self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout
        sys.stderr = self._stderr


def skip_if_mysql(fn):
    if connection.vendor == "mysql":
        return unittest.skip("MySQL does not support tested feature")(fn)
    return fn



"""
Test models
"""
from django.db import models

import tagulous


class MixedModel(models.Model):
    """
    For testing mixed model in a separate app
    """

    name = models.CharField(blank=True, max_length=100)
    singletag = tagulous.models.SingleTagField(
        blank=True, null=True, initial="Mr, Mrs, Ms"
    )
    tags = tagulous.models.TagField(blank=True, initial="Adam, Brian, Chris")



"""
Tagulous test app urls

Usage:
    class MyTestCase(TestCase):
        urls = 'tests.tagulous_tests_app.urls'
"""
from django.conf.urls import include, re_path

import tagulous.views
from tests.tagulous_tests_app import models, views


tagged_model = models.TagFieldOptionsModel

urlpatterns = [
    # We don't include the admin site here - that's added by AdminTestManager
    re_path(
        r"^tagulous_tests_app/",
        include(
            [
                # CBV with tagged forms
                re_path(r"views/$", views.null, name="tagulous_tests_app-null"),
                re_path(
                    r"views/MixedCreate/$",
                    views.MixedCreate.as_view(),
                    name="tagulous_tests_app-MixedCreate",
                ),
                re_path(
                    r"views/MixedUpdate/(?P<pk>[0-9]+)/$",
                    views.MixedUpdate.as_view(),
                    name="tagulous_tests_app-MixedUpdate",
                ),
                # Tagulous autocomplete
                re_path(
                    r"^autocomplete/unlimited/$",
                    tagulous.views.autocomplete,
                    {"tag_model": tagged_model.autocomplete_view.tag_model},
                    name="tagulous_tests_app-unlimited",
                ),
                re_path(
                    r"^autocomplete/limited/$",
                    tagulous.views.autocomplete,
                    {"tag_model": tagged_model.autocomplete_limit.tag_model},
                    name="tagulous_tests_app-limited",
                ),
                re_path(
                    r"^autocomplete/unlimited/login/$",
                    tagulous.views.autocomplete_login,
                    {"tag_model": tagged_model.autocomplete_view.tag_model},
                    name="tagulous_tests_app-login",
                ),
                re_path(
                    r"^autocomplete/unlimited/queryset/$",
                    tagulous.views.autocomplete,
                    {
                        "tag_model": tagged_model.autocomplete_view.tag_model.objects.filter(
                            name__startswith="tag2"
                        )
                    },
                    name="tagulous_tests_app-queryset",
                ),
                re_path(
                    r"^autocomplete/unlimited/force_lowercase_true/$",
                    tagulous.views.autocomplete,
                    {"tag_model": tagged_model.force_lowercase_true.tag_model},
                    name="tagulous_tests_app-force_lowercase_true",
                ),
                re_path(
                    r"^autocomplete/unlimited/case_sensitive_false/$",
                    tagulous.views.autocomplete,
                    {"tag_model": tagged_model.case_sensitive_false.tag_model},
                    name="tagulous_tests_app-case_sensitive_false",
                ),
                re_path(
                    r"^autocomplete/unlimited/case_sensitive_true/$",
                    tagulous.views.autocomplete,
                    {"tag_model": tagged_model.case_sensitive_true.tag_model},
                    name="tagulous_tests_app-case_sensitive_true",
                ),
            ]
        ),
    ),
]



# -*- coding: utf-8 -*-
from django.db import migrations, models

import tagulous.models.fields
import tagulous.models.migrations


class Migration(migrations.Migration):

    dependencies = [("tagulous_tests_migration", "0002_tagged")]

    operations = (
        [
            migrations.AddField(
                model_name="tagulous_migrationtestmodel_tags",
                name="parent",
                field=models.ForeignKey(
                    to="tagulous_tests_migration.Tagulous_MigrationTestModel_tags",
                    related_name="children",
                    blank=True,
                    null=True,
                    on_delete=models.CASCADE,
                ),
                preserve_default=True,
            ),
            migrations.AddField(
                model_name="tagulous_migrationtestmodel_tags",
                name="label",
                field=models.CharField(
                    default="-",
                    max_length=191,
                    help_text=b"The name of the tag, without ancestors",
                ),
                preserve_default=True,
            ),
            migrations.AddField(
                model_name="tagulous_migrationtestmodel_tags",
                name="level",
                field=models.IntegerField(
                    default=1, help_text=b"The level of the tag in the tree"
                ),
                preserve_default=True,
            ),
        ]
        + tagulous.models.migrations.add_unique_field(
            model_name="tagulous_migrationtestmodel_tags",
            name="path",
            field=models.TextField(),
            preserve_default=False,
            set_fn=lambda obj: setattr(obj, "path", str(obj.pk)),
        )
        + [
            migrations.AlterField(
                model_name="migrationtestmodel",
                name="tags",
                field=tagulous.models.fields.TagField(
                    to="tagulous_tests_migration.Tagulous_MigrationTestModel_tags",
                    help_text=b"Enter a comma-separated tag string",
                    _set_tag_meta=True,
                    tree=True,
                ),
                preserve_default=True,
            ),
            migrations.AlterUniqueTogether(
                name="tagulous_migrationtestmodel_tags",
                unique_together=set([("slug", "parent")]),
            ),
            tagulous.models.migrations.ChangeModelBases(
                name="tagulous_migrationtestmodel_tags",
                bases=(tagulous.models.models.BaseTagTreeModel, models.Model),
            ),
        ]
    )



"""
Tagulous test: django migrations

Modules tested:
    tagulous.models.migrations
"""
import os
import shutil
import sys
import warnings
from importlib import import_module

from django.core.management import call_command
from django.db import DatabaseError
from django.test import TransactionTestCase

from tagulous import models as tag_models
from tests import tagulous_tests_migration
from tests.lib import Capturing, TagTestManager, testenv


# If True, display output from call_command - use for debugging tests
DISPLAY_CALL_COMMAND = False

app_name = "tagulous_tests_migration"
app_module = sys.modules["tests.%s" % app_name]
migrations_name = "migrations_%s" % testenv
migrations_module = "tests.%s.%s" % (app_name, migrations_name)
migrations_path = None


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Util functions
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


def clear_migrations():
    "Clear cached mentions of django migrations to force a reload"
    # Remove loaded migrations
    if hasattr(app_module, migrations_name):
        delattr(app_module, migrations_name)

    for key in list(sys.modules.keys()):
        if key.startswith(migrations_module):
            del sys.modules[key]

    try:
        from importlib import invalidate_caches
    except ImportError:
        pass
    else:
        invalidate_caches()


def get_migrations():
    """
    Clears migration cache and gets migrations for the test app which have not
    yet been applied
    """
    from django.db.migrations.loader import MigrationLoader

    clear_migrations()
    loader = MigrationLoader(None, ignore_no_migrations=True)
    graph = loader.graph
    return graph.leaf_nodes(app_name)


def get_migrations_dir():
    "Get migration dir"
    if migrations_path is None:
        globals()["migrations_path"] = os.path.join(
            os.path.dirname(tagulous_tests_migration.__file__), migrations_name
        )
    return migrations_path


def get_expected_dir():
    return os.path.normpath(
        os.path.join(get_migrations_dir(), "..", "django_migrations_expected")
    )


def clean_all():
    "Clean everything - roll back and delete any migrations, forget any loaded"
    # Delete migrations
    migrations_dir = get_migrations_dir()
    expected_dir = get_expected_dir()
    if not (app_name in migrations_dir and migrations_dir.endswith(migrations_name)):
        # Catch unexpected path - don't want to delete anything important
        raise ValueError("Migrations dir has unexpected name: %s" % migrations_dir)
    if os.path.isdir(migrations_dir):
        shutil.rmtree(migrations_dir)
    elif os.path.exists(migrations_dir):
        raise ValueError("Migrations dir is not a dir")

    # Try to roll back to zero using expected migrations
    shutil.copytree(expected_dir, migrations_dir)

    try:
        migrate_app(target="zero")
    except DatabaseError:
        # Guess it didn't exist - that's ok, nothing to reverse
        pass
    shutil.rmtree(migrations_dir)

    # Empty models
    tagulous_tests_migration.models.unset_model()

    # Clear django's migration cache
    clear_migrations()


def make_migration(name):
    "Make a migration with the given name"
    try:
        with Capturing() as output:
            call_command(
                "makemigrations",
                "--name={}".format(name),
                app_name,  # app to make migration for
                verbosity=0,  # Silent
            )
    except Exception as e:
        print(">> makemigration failed for {}:".format(name))
        print("\n".join(output))
        print("<<<<<<<<<<")
        raise e

    clear_migrations()


def migrate_app(target=None):
    "Apply migrations"
    clear_migrations()

    if DISPLAY_CALL_COMMAND:
        print(">> manage.py migrate %s target=%s" % (app_name, target))

    args = [app_name]
    if target is not None:
        args.append(target)
    kwargs = {"interactive": False, "verbosity": 1}  # no user input  # basic reporting

    try:
        with Capturing() as output:
            with warnings.catch_warnings(record=True) as cw:
                call_command("migrate", *args, **kwargs)
    except Exception as e:
        print(">> migration failed: %s" % e)
        if not DISPLAY_CALL_COMMAND:
            print(">> manage.py migrate %s target=%s" % (app_name, target))
        print("\n".join(output))
        print("<<<<<<<<<<")
        raise

    # Ensure caught warnings are expected
    for w in cw:
        assert issubclass(w.category, PendingDeprecationWarning)

    if DISPLAY_CALL_COMMAND:
        print("\n".join(output))
        print("<<<<<<<<<<")

    return output


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Tests
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class DjangoMigrationTest(TagTestManager, TransactionTestCase):
    """
    Test django migrations
    """

    #
    # Test management - ensure it's clean before and after
    #

    @classmethod
    def setUpClass(self):
        "Clean everything before each test in case previous run failed"
        clean_all()

    def tearDownExtra(self):
        "Clean away the model so it's not installed by TestCase"
        clean_all()

    @classmethod
    def tearDownClass(cls):
        "Leave everything clean at the end of the tests"
        clean_all()

    #
    # Migration file analysis
    #

    def _import_migration(self, name):
        "Import the named migration"
        path = get_migrations_dir()
        root = os.path.dirname(__file__)
        if not path.startswith(root):
            return self.fail(
                "Could not find common root between test and migration dir:\n"
                "  %s\n  %s" % __file__,
                path,
            )

        module_name = ".".join(["tests"] + path[len(root) + 1 :].split(os.sep) + [name])
        module = import_module(module_name)
        return module

    #
    # Extra assertions
    #

    def assertMigrationExpected(self, name):
        "Compare two migration files"
        # Import and validate the migrations
        mod1 = self._import_migration(name)
        self.assertValidMigration(mod1)
        mig1 = mod1.Migration

        mod2 = self._import_migration(name)
        self.assertValidMigration(mod2)
        mig2 = mod2.Migration

        # Compare dependencies
        # Straight list of tuples
        self.assertEqual(mig1.dependencies, mig2.dependencies)

        # Compare operations
        # They're list of Operation classes, which provide __eq__
        self.assertEqual(mig1.operations, mig2.operations)

    def assertValidMigration(self, module):
        "Basic validation of an imported migration"
        from django.db.migrations.migration import Migration

        self.assertTrue("Migration" in module.__dict__)
        self.assertTrue(issubclass(module.Migration, Migration))
        self.assertTrue(hasattr(module.Migration, "dependencies"))
        self.assertTrue(hasattr(module.Migration, "operations"))

    #
    # Tests
    #

    def migrate_initial(self):
        """
        Load the initial test app's model, and create and apply an initial
        schema migration. Check that the migration worked.
        """
        model_initial = tagulous_tests_migration.models.set_model_initial()
        self.assertFalse(issubclass(model_initial, tag_models.tagged.TaggedModel))

        # Make the migration
        make_migration("initial")

        # Check the files were created as expected
        migrations = get_migrations()
        self.assertEqual(len(migrations), 1)
        self.assertEqual(migrations[0], (app_name, "0001_initial"))
        self.assertMigrationExpected("0001_initial")

        # Check they apply correctly
        output = migrate_app()

        self.assertEqual(
            output,
            [
                "Operations to perform:",
                "  Apply all migrations: tagulous_tests_migration",
                "Running migrations:",
            ]
            + ["  Applying tagulous_tests_migration.0001_initial... OK"],
        )

        return model_initial

    def migrate_tagged(self):
        """
        After migrating to the initial model, switch the test app to the
        tagged model, and create and apply a schema migration. Check that the
        migration worked.
        """
        # First need to migrate to initial - re-run that migration
        self.migrate_initial()

        # Now switch model
        model_tagged = tagulous_tests_migration.models.set_model_tagged()
        self.assertTrue(issubclass(model_tagged, tag_models.tagged.TaggedModel))

        # Make the migration
        make_migration("tagged")

        # Check the files were created as expected
        migrations = get_migrations()
        self.assertEqual(len(migrations), 1)
        self.assertEqual(migrations[0], (app_name, "0002_tagged"))
        self.assertMigrationExpected("0002_tagged")

        # Check they apply correctly
        output = migrate_app()
        self.assertEqual(
            output,
            [
                "Operations to perform:",
                "  Apply all migrations: tagulous_tests_migration",
                "Running migrations:",
            ]
            + ["  Applying tagulous_tests_migration.0002_tagged... OK"],
        )

        return model_tagged

    def migrate_tree(self):
        """
        After migrating to the tagged model, switch the test app to the tree
        model, load some test data, and apply the pre-written schema migration
        which uses add_unique_column. Check that the migration worked.
        """
        # Migrate to tagged
        model_tagged = self.migrate_tagged()

        # Add some test data
        model_tagged.tags.tag_model.objects.create(name="one/two/three")
        model_tagged.tags.tag_model.objects.create(name="uno/dos/tres")
        self.assertTagModel(
            model_tagged.tags.tag_model, {"one/two/three": 0, "uno/dos/tres": 0}
        )

        # Now switch model
        model_tree = tagulous_tests_migration.models.set_model_tree()
        self.assertTrue(issubclass(model_tree, tag_models.tagged.TaggedModel))
        self.assertTagModel(
            model_tagged.tags.tag_model, {"one/two/three": 0, "uno/dos/tres": 0}
        )

        # We can't automaticallly create a migration here; because we're adding
        # a null field, makemigration would ask us questions we don't want to
        # answer anyway, because we'd replace the AddField operation with a
        # tagulous.models.mgiration.AddUniqueField operation. We'll therefore
        # use one we prepared earlier, 0003_tree.py

        # Add in the prepared schemamigration for the tree
        migrations_dir = get_migrations_dir()
        expected_dir = get_expected_dir()
        shutil.copy(os.path.join(expected_dir, "0003_tree.py"), migrations_dir)

        # Check the files were created as expected
        migrations = get_migrations()
        self.assertEqual(len(migrations), 1)
        self.assertEqual(migrations[0], (app_name, "0003_tree"))

        # Check they apply correctly
        output = migrate_app()
        self.assertSequenceEqual(
            output,
            [
                "Operations to perform:",
                "  Apply all migrations: tagulous_tests_migration",
                "Running migrations:",
            ]
            + ["  Applying tagulous_tests_migration.0003_tree... OK"],
        )

        # Data shouldn't have changed yet
        self.assertTagModel(
            model_tree.tags.tag_model, {"one/two/three": 0, "uno/dos/tres": 0}
        )
        self.assertEqual(model_tree.tags.tag_model.objects.get(pk=1).path, "1")
        self.assertEqual(model_tree.tags.tag_model.objects.get(pk=2).path, "2")

        # Rebuild tree
        model_tree.tags.tag_model.objects.rebuild()

        # We should now have nicely-built trees
        self.assertTagModel(
            model_tree.tags.tag_model,
            {
                "one": 0,
                "one/two": 0,
                "one/two/three": 0,
                "uno": 0,
                "uno/dos": 0,
                "uno/dos/tres": 0,
            },
        )
        tag_objects = model_tree.tags.tag_model.objects
        self.assertEqual(tag_objects.get(name="one").path, "one")
        self.assertEqual(tag_objects.get(name="one/two").path, "one/two")
        self.assertEqual(tag_objects.get(name="one/two/three").path, "one/two/three")
        self.assertEqual(tag_objects.get(name="uno").path, "uno")
        self.assertEqual(tag_objects.get(name="uno/dos").path, "uno/dos")
        self.assertEqual(tag_objects.get(name="uno/dos/tres").path, "uno/dos/tres")

        return model_tree

    def migrate_data(self):
        """
        After migrating to the tree model, apply the pre-written data migration
        which tests tag fields and models. Check that the migration worked.
        """
        model_tree = self.migrate_tree()

        # Empty the tags from the test model added by migrate_tree
        model_tree.tags.tag_model.objects.all().delete()
        self.assertTagModel(model_tree.tags.tag_model, {})

        # Add some test data to the model itself
        model_tree.objects.create(
            name="Test 1", singletag="Mr", tags="one/two, uno/dos"
        )
        model_tree.objects.create(name="Test 2", singletag="Mrs", tags="one/two")
        model_tree.objects.create(name="Test 3", singletag="Mr", tags="uno/dos")
        self.assertTagModel(model_tree.singletag.tag_model, {"Mr": 2, "Mrs": 1})
        self.assertTagModel(
            model_tree.tags.tag_model, {"one": 0, "one/two": 2, "uno": 0, "uno/dos": 2}
        )

        # Add in the datamigration
        migrations_dir = get_migrations_dir()
        expected_dir = get_expected_dir()
        shutil.copy(os.path.join(expected_dir, "0004_data.py"), migrations_dir)

        # Check the files were created as expected
        migrations = get_migrations()
        self.assertEqual(len(migrations), 1)
        self.assertEqual(migrations[0], (app_name, "0004_data"))

        # Check they apply correctly
        output = migrate_app()
        self.assertSequenceEqual(
            output,
            [
                "Operations to perform:",
                "  Apply all migrations: tagulous_tests_migration",
                "Running migrations:",
            ]
            + ["  Applying tagulous_tests_migration.0004_data... OK"],
        )

    #
    # Tests
    #

    # Individual tests for development purposes
    # No point running each of them - test_data() will run them as its setup
    """
    def test_initial(self):
        "Test initial migration is created and can be applied and used"
        self.migrate_initial()

    def test_tagged(self):
        "Test tagged migration is created and can be applied and used"
        self.migrate_tagged()

    def test_tree(self):
        "Test migration to Tree model using add_unique_column"
        self.migrate_tree()
    """

    def test_data(self):
        "Test data migration"
        self.migrate_data()



from .fields import SingleTagField, TagField


#
# Functions to load initial tags
# Used by tests and the management command `initialtags`
#


def field_initialise_tags(model, field, report=None):
    """
    Load any initial tags for the specified tag field

    You will not normally need to call this directly - instead use the
    management command ``initialtags``.

    Arguments:
        model       Model containing the field
        field       Field with initial tags to load
        report      Optional: a file handle to write verbose reports to

    Returns True if loaded, False if nothing to load
    """
    if not field.tag_options.initial:
        return False

    if report:
        report.write(
            "Loading initial tags for %s.%s.%s\n"
            % (model._meta.app_label, model.__name__, field.name)
        )

    descriptor = getattr(model, field.name)
    descriptor.load_initial()
    return True


def model_initialise_tags(model, report=None):
    """
    Load any initial tags for the given model

    You will not normally need to call this directly - instead use the
    management command ``initialtags``.

    Arguments:
        model       Model to check for tag fields to load
        report      Optional: a file handle to write verbose reports to
    """
    fields = model._meta.get_fields()
    for field in fields:
        if isinstance(field, (SingleTagField, TagField)):
            field_initialise_tags(model, field, report)



import json

from django.contrib.auth.decorators import login_required
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.query import QuerySet
from django.http import HttpResponse


@login_required
def autocomplete_login(*args, **kwargs):
    return autocomplete(*args, **kwargs)


def autocomplete(request, tag_model):
    """
    Arguments:
        request
            The request object from the dispatcher
        tag_model
            Reference to the tag model (eg MyModel.tags.tag_model), or a
            queryset of the tag model (eg MyModel.tags.tag_model.objects.all())

    The following GET parameters can be set:
        q   The query string to filter by (match against start of string)
        p   The current page

    Response is a JSON object with following keys:
        results     List of tags
        more        Boolean if there is more
    }
    """
    # Get model, queryset and tag options
    if isinstance(tag_model, QuerySet):
        queryset = tag_model
        tag_model = queryset.model
    else:
        queryset = tag_model.objects
    options = tag_model.tag_options

    # Get query string
    query = request.GET.get("q", "")
    page = int(request.GET.get("p", 1))

    # Perform search
    if query:
        if options.force_lowercase:
            query = query.lower()

        if options.autocomplete_view_fulltext:
            lookup = "contains"
        else:
            lookup = "startswith"

        if not options.case_sensitive:
            lookup = f"i{lookup}"

        results = queryset.filter(**{f"name__{lookup}": query})

    else:
        results = queryset.all()

    # Limit results
    if options.autocomplete_limit:
        start = options.autocomplete_limit * (page - 1)
        end = options.autocomplete_limit * page
        more = results.count() > end
        results = results.order_by("name")[start:end]

    # Build response
    response = {"results": [tag.name for tag in results], "more": more}
    return HttpResponse(
        json.dumps(response, cls=DjangoJSONEncoder), content_type="application/json"
    )



"""
Tagulous test: manage.py commands

Modules tested:
    tagulous.management.commands.initial_tags
"""
from django.core.management import call_command
from django.test import TestCase

from tests.lib import Capturing, TagTestManager
from tests.tagulous_tests_app import models as test_models
from tests.tagulous_tests_app2 import models as test_models2


# If True, display output from call_command - use for debugging tests
DISPLAY_CALL_COMMAND = False


# ##############################################################################
# ###### ./manage.py initial_tags
# ##############################################################################


class InitialTagsTest(TagTestManager, TestCase):
    """
    Test initial_tags command
    """

    # Don't list the models in manage_models, otherwise TagTestManager will
    # initialise the tags for us
    def setUpExtra(self):
        self.singletag_model = test_models.SingleTagFieldOptionsModel.initial_list
        self.singletag_model_alt = test_models.SingleTagFieldOptionsModel.initial_string
        self.tag_model = test_models.TagFieldOptionsModel.initial_list
        self.singletag_model_2 = test_models2.MixedModel.singletag
        self.tag_model_2 = test_models2.MixedModel.tags

    def assertModelsEmpty(self):
        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.singletag_model_alt, {})
        self.assertTagModel(self.tag_model, {})
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def assertSingleTagFilled(self, model):
        self.assertTagModel(model, {"Mr": 0, "Mrs": 0, "Ms": 0})

    def assertTagFilled(self, model):
        self.assertTagModel(model, {"Adam": 0, "Brian": 0, "Chris": 0})

    def run_command(self, target=""):
        with Capturing() as output:
            call_command(
                "initial_tags", target=target, verbosity=1  # Optional target  # Silent
            )

        if DISPLAY_CALL_COMMAND:
            print(">> initial_tags target=%s" % target)
            print("\n".join(output))
            print("<<<<<<<<<<")

        return output

    def test_load_all(self):
        "Check no target loads all"
        self.assertModelsEmpty()
        output = self.run_command()

        # Output correct if at least expected fields are loaded (more in apps)
        expected = [
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_list",
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_string",
            "tagulous_tests_app.TagFieldOptionsModel.initial_list",
            "tagulous_tests_app2.MixedModel.singletag",
            "tagulous_tests_app2.MixedModel.tags",
        ]
        self.assertGreaterEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue("Loading initial tags for %s" % field in output)

        # All loaded
        self.assertSingleTagFilled(self.singletag_model)
        self.assertSingleTagFilled(self.singletag_model_alt)
        self.assertTagFilled(self.tag_model)
        self.assertSingleTagFilled(self.singletag_model_2)
        self.assertTagFilled(self.tag_model_2)

    def test_load_app(self):
        "Check app target just loads that app"
        self.assertModelsEmpty()
        output = self.run_command("tagulous_tests_app")

        # Output correct if at least expected fields are loaded (more in apps)
        expected = [
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_list",
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_string",
            "tagulous_tests_app.TagFieldOptionsModel.initial_list",
        ]
        self.assertGreaterEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue("Loading initial tags for %s" % field in output)

        # Test app 1 loaded
        self.assertSingleTagFilled(self.singletag_model)
        self.assertSingleTagFilled(self.singletag_model_alt)
        self.assertTagFilled(self.tag_model)

        # Test app 2 empty
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def test_load_model(self):
        "Check app and model target just loads that model"
        self.assertModelsEmpty()
        output = self.run_command("tagulous_tests_app.SingleTagFieldOptionsModel")

        # Output correct if at least expected fields are loaded (more in model)
        expected = [
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_list",
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_string",
        ]
        self.assertGreaterEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue("Loading initial tags for %s" % field in output)

        # SingleTagFieldOptionsModel loaded
        self.assertSingleTagFilled(self.singletag_model)
        self.assertSingleTagFilled(self.singletag_model_alt)

        # Rest not loaded
        self.assertTagModel(self.tag_model, {})
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def test_load_field(self):
        "Check app, model and field target just loads that field"
        self.assertModelsEmpty()
        output = self.run_command(
            "tagulous_tests_app.SingleTagFieldOptionsModel.initial_list"
        )

        # Output correct if only expected fields are loaded
        expected = ["tagulous_tests_app.SingleTagFieldOptionsModel.initial_list"]
        self.assertEqual(len(output), len(expected))
        for field in expected:
            self.assertTrue("Loading initial tags for %s" % field in output)

        # SingleTagFieldOptionsModel loaded
        self.assertSingleTagFilled(self.singletag_model)

        # Rest not loaded
        self.assertTagModel(self.singletag_model_alt, {})
        self.assertTagModel(self.tag_model, {})
        self.assertTagModel(self.singletag_model_2, {})
        self.assertTagModel(self.tag_model_2, {})

    def test_load_field_empty(self):
        "Check loading a field which is empty prints the message"
        self.assertModelsEmpty()
        field = "tagulous_tests_app.SingleTagFieldOptionsModel.protect_all_true"
        output = self.run_command(field)
        self.assertSequenceEqual(output, ["Nothing to load for %s" % field])
        self.assertModelsEmpty()



"""
Tagulous constants
"""
# Prefix for all models auto-generated by Tagulous
MODEL_PREFIX = "Tagulous"

# Attribute of manager on the tagged model
# %s will be replaced by the field name
TAGGED_ATTR_MANAGER = "_%s_tagulous"

# Constants to improve code legibility
COMMA = ","
SPACE = " "
QUOTE = '"'
DOUBLE_QUOTE = QUOTE + QUOTE
TREE = "/"

# Default model TagField options
OPTION_DEFAULTS = {
    "initial": "",
    "protect_initial": True,
    "protect_all": False,
    "case_sensitive": False,
    "force_lowercase": False,
    "max_count": 0,
    "space_delimiter": True,
    "tree": False,
    "autocomplete_initial": False,
    "autocomplete_view": "",
    "autocomplete_view_args": None,
    "autocomplete_view_kwargs": None,
    "autocomplete_view_fulltext": False,
    "autocomplete_limit": 100,
    "autocomplete_settings": None,
    "get_absolute_url": None,
    "verbose_name_singular": None,
    "verbose_name_plural": None,
}

# List of model TagField options which are relevant to form fields
FORM_OPTIONS = [
    "case_sensitive",
    "force_lowercase",
    "max_count",
    "space_delimiter",
    "tree",
    "autocomplete_limit",
    "autocomplete_settings",
]



"""
Tagulous test: Form SingleTagField

Modules tested:
    tagulous.models.fields.SingleTagField
    tagulous.forms.SingleTagField
"""
from django import forms
from django.test import TestCase

from tagulous import forms as tag_forms
from tagulous import models as tag_models
from tagulous import settings as tag_settings
from tests.lib import TagTestManager, skip_if_mysql
from tests.tagulous_tests_app import forms as test_forms
from tests.tagulous_tests_app import models as test_models


# ##############################################################################
# ###### Test form SingleTagField
# ##############################################################################


class FormSingleTagFieldTest(TagTestManager, TestCase):
    """
    Test form SingleTagFields being used directly (without ModelForm)
    """

    def test_required(self):
        "Test required status is passed from field to widget"
        self.assertTrue(tag_forms.SingleTagField(required=True).required)
        self.assertTrue(tag_forms.SingleTagField(required=True).widget.is_required)
        self.assertFalse(tag_forms.SingleTagField(required=False).required)
        self.assertFalse(tag_forms.SingleTagField(required=False).widget.is_required)
        self.assertTrue(tag_forms.SingleTagField().required)
        self.assertTrue(tag_forms.SingleTagField().widget.is_required)

    def test_input(self):
        "Test valid and invalid input"
        self.assertFieldOutput(
            tag_forms.SingleTagField,
            valid={"Mr": "Mr", "Mr, Mrs": "Mr, Mrs"},
            invalid={},
            empty_value=None,
        )

    def test_force_lowercase(self):
        "Test force_lowercase flag set in constructor"
        self.assertFieldOutput(
            tag_forms.SingleTagField,
            field_kwargs={"tag_options": tag_models.TagOptions(force_lowercase=True)},
            valid={"Mr": "mr", "Mr, Mrs": "mr, mrs"},
            invalid={},
            empty_value=None,
        )

    def test_response(self):
        "Test response is as expected"
        form = test_forms.SingleTagFieldForm(data={"singletag": "Mr"})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["singletag"], "Mr")

    def test_whitespace_in_required__fails(self):
        "Test whitespace is stripped and required field fails"
        form = test_forms.SingleTagFieldForm(data={"singletag": " "})
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {"singletag": ["This field is required."]})

    def test_whitespace_in_optional__is_none(self):
        "Test whitespace is stripped and optional field returns None"

        class LocalTestForm(forms.Form):
            tag = tag_forms.SingleTagField(required=False)

        form = LocalTestForm(data={"tag": " "})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["tag"], None)

    def test_render_basics(self):
        "Check widget renders default settings (not required)"

        class LocalTestForm(forms.Form):
            tag = tag_forms.SingleTagField()

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: true, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'id="id_tag" name="tag" {{required}}type="text" />'
            ),
        )

    def test_render_tag_optional(self):
        "Check widget renders correctly when field not required"

        class LocalTestForm(forms.Form):
            tag = tag_forms.SingleTagField(required=False)

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: false, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'id="id_tag" name="tag" type="text" />'
            ),
        )

    def test_render_tag_list(self):
        "Check widget renders data-tag-list"

        class LocalTestForm(forms.Form):
            tag = tag_forms.SingleTagField(autocomplete_tags=["one", "two", "three"])

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: true, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'data-tag-list="'
                '[&quot;one&quot;, &quot;two&quot;, &quot;three&quot;]" '
                'id="id_tag" name="tag" {{required}}type="text" />'
            ),
        )

    def test_render_tag_url(self):
        "Check widget renders data-tag-url"
        autocomplete_view = "tagulous_tests_app-null"

        class LocalTestForm(forms.Form):
            tag = tag_forms.SingleTagField(
                tag_options=tag_models.TagOptions(autocomplete_view=autocomplete_view)
            )

        form = LocalTestForm()
        self.assertHTMLEqual(
            str(form["tag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: true, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'data-tag-url="'
                '/tagulous_tests_app/views/" '
                'id="id_tag" name="tag" {{required}}type="text" />'
            ),
        )

    def test_render_value(self):
        "Check widget renders value"
        form = test_forms.SingleTagFieldForm(data={"singletag": "Mr"})
        self.assertHTMLEqual(
            str(form["singletag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: true, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'id="id_singletag" name="singletag" {{required}}type="text" '
                'value="Mr" />'
            ),
        )


# ##############################################################################
# ###### Test ModelForm SingleTagField
# ##############################################################################


class ModelFormSingleTagFieldTest(TagTestManager, TestCase):
    """
    Test modelform SingleTagFields
    """

    manage_models = [test_models.SingleTagFieldModel]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.form = test_forms.SingleTagFieldModelForm
        self.model = test_models.SingleTagFieldModel
        self.tag_model = self.model.title.tag_model

    def test_formfield(self):
        "Test that model.SingleTagField.formfield works correctly"
        # Check that the model fields are generated correctly
        tag1_field = self.model._meta.get_field("title").formfield()
        self.assertIsInstance(tag1_field, tag_forms.SingleTagField)
        self.assertIsInstance(tag1_field.tag_options, tag_models.TagOptions)

        # Check we can get it from the descriptor shortcut method
        tag2_field = self.model.title.formfield()
        self.assertIsInstance(tag2_field, tag_forms.SingleTagField)
        self.assertIsInstance(tag2_field.tag_options, tag_models.TagOptions)

    def test_media(self):
        "Test that a model form with a SingleTagField has the correct media"
        media = self.form().media
        for js in tag_settings.AUTOCOMPLETE_JS:
            self.assertTrue(js in media._js)
        for grp, files in tag_settings.AUTOCOMPLETE_CSS.items():
            self.assertTrue(grp in media._css)
            for css in files:
                self.assertTrue(css in media._css[grp])

    def test_single_model_form_save(self):
        """
        Test that a model form with a SingleTagField saves correctly
        """
        form = self.form(data={"name": "Test 1", "title": "Mr"})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.title, "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_override_option_dict_formfield(self):
        "Test defaults can be overriden in formfield() call with a dict"
        # Confirm default
        field1 = self.form().fields["title"]
        self.assertEqual(field1.tag_options.case_sensitive, False)

        # Change default
        field2 = self.model.title.formfield(tag_options={"case_sensitive": True})
        self.assertEqual(field2.tag_options.case_sensitive, True)

    def test_override_option_cls_formfield(self):
        "Test defaults can be overriden in formfield() call with a TagOption"
        # Confirm default
        field1 = self.form().fields["title"]
        self.assertEqual(field1.tag_options.case_sensitive, False)

        # Change default
        field2 = self.model.title.formfield(
            tag_options=tag_models.TagOptions(case_sensitive=True)
        )
        self.assertEqual(field2.tag_options.case_sensitive, True)

    def test_override_autocomplete_tags_formfield(self):
        "Test list of autocomplete tags can be passed in formfield"
        self.tag_model.objects.create(name="Mr")
        self.tag_model.objects.create(name="Mrs")

        # Confirm default
        field1 = self.form().fields["title"]
        self.assertSequenceEqual(
            [t.name for t in field1.autocomplete_tags],
            [t.name for t in self.tag_model.objects.all()],
        )

        # Change default
        field2 = self.model.title.formfield(autocomplete_tags=["Ms", "Mx"])
        self.assertSequenceEqual(field2.autocomplete_tags, ["Ms", "Mx"])

    def test_render_tag_list(self):
        "Check widget renders data-tag-list"
        self.tag_model.objects.create(name="Mr")
        self.tag_model.objects.create(name="Mrs")
        self.assertTagModel(self.tag_model, {"Mr": 0, "Mrs": 0})
        form = self.form(data={"name": "Test 1", "title": "Mrs"})
        self.assertHTMLEqual(
            str(form["title"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: false, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'data-tag-list="[&quot;Mr&quot;, &quot;Mrs&quot;]" '
                'id="id_title" name="title" type="text" value="Mrs" />'
            ),
        )

    def test_initial_string(self):
        "Check initial string"
        form = test_forms.SingleTagFieldForm(initial={"singletag": "Mr"})
        self.assertHTMLEqual(
            str(form["singletag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: true, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'id="id_singletag" name="singletag" {{required}}type="text" '
                'value="Mr" />'
            ),
        )

    def test_initial_tag(self):
        "Check initial tag"
        t1 = self.tag_model.objects.create(name="Mr")
        form = test_forms.SingleTagFieldForm(initial={"singletag": t1})
        self.assertHTMLEqual(
            str(form["singletag"]),
            (
                '<input autocomplete="off" '
                'data-tag-options="{'
                '&quot;required&quot;: true, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'id="id_singletag" name="singletag" {{required}}type="text" '
                'value="Mr" />'
            ),
        )

    def test_tagged_edit(self):
        "Check edit tagged model form instance works"
        t1 = self.model.objects.create(name="Test 1", title="Mr")
        form = self.form(instance=t1)
        self.assertHTMLEqual(
            str(form["title"]),
            (
                '<input autocomplete="off" '
                'data-tag-list="[&quot;Mr&quot;]" '
                'data-tag-options="{'
                '&quot;required&quot;: false, &quot;max_count&quot;: 1}" '
                'data-tag-type="single" data-tagulous="true" '
                'id="id_title" name="title" type="text" '
                'value="Mr" />'
            ),
        )

    def test_tag_with_delims(self):
        "Check tag with delimiters"
        t1 = self.model.objects.create(name="Test 1", title="One, Two")
        form = self.form(instance=t1, data={"name": t1.name, "title": t1.title})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["title"], "One, Two")
        t2 = form.save()
        self.assertEqual(t2.name, "Test 1")
        self.assertEqual(t2.title, "One, Two")

    def test_tag_with_quotes(self):
        "Check tag with quotes"
        t1 = self.model.objects.create(name="Test 1", title='"One, "Two"')
        form = self.form(instance=t1, data={"name": t1.name, "title": t1.title})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["title"], '"One, "Two"')
        t2 = form.save()
        self.assertEqual(t2.name, "Test 1")
        self.assertEqual(t2.title, '"One, "Two"')


# ##############################################################################
# ######  Test SingleTagField blank
# ##############################################################################


class ModelFormSingleTagFieldOptionalTest(TagTestManager, TestCase):
    "Test optional SingleTagField"
    manage_models = [test_models.SingleTagFieldOptionalModel]

    def setUpExtra(self):
        self.form = test_forms.SingleTagFieldOptionalModelForm
        self.model = test_models.SingleTagFieldOptionalModel
        self.tag_model = self.model.tag.tag_model

    def test_optional_singletagfield(self):
        "Check model with optional singletagfield can be saved when empty"
        form = self.form(data={"name": "Test 1"})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tag, None)
        self.assertTagModel(self.tag_model, {})


class ModelFormSingleTagFieldRequiredTest(TagTestManager, TestCase):
    "Test required SingleTagField"
    manage_models = [test_models.SingleTagFieldRequiredModel]

    def setUpExtra(self):
        self.form = test_forms.SingleTagFieldRequiredModelForm
        self.model = test_models.SingleTagFieldRequiredModel
        self.tag_model = self.model.tag.tag_model

    def test_required_singletagfield_set(self):
        "Check model with required singletagfield can be saved when set"
        form = self.form(data={"name": "Test 1", "tag": "Mr"})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, "Test 1")
        self.assertEqual(t1.tag, "Mr")
        self.assertTagModel(self.tag_model, {"Mr": 1})

    def test_required_singletagfield_empty(self):
        "Check model with required singletagfield cannot be saved when empty"
        form = self.form(data={"name": "Test 1"})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], "tag")
        self.assertEqual(len(form.errors["tag"]), 1)
        self.assertEqual(form.errors["tag"][0], "This field is required.")


# ##############################################################################
# ###### Test form SingleTagField options
# ##############################################################################


class ModelFormSingleTagFieldOptionsTest(TagTestManager, TestCase):
    "Check tag options"
    manage_models = [test_models.SingleTagFieldOptionsModel]

    def setUpExtra(self):
        self.form = test_forms.SingleTagFieldOptionsModelForm
        self.model = test_models.SingleTagFieldOptionsModel

    @skip_if_mysql
    def test_case_sensitive_true(self):
        "Test form SingleTagField case_sensitive true"
        # Prep tag model
        tag_model = self.model.case_sensitive_true.tag_model
        tag_model.objects.get_or_create(name="Mr")

        # Check that the option is passed to the form field from model field
        form = self.form(data={"case_sensitive_true": "mr"})
        field = form.fields["case_sensitive_true"]
        self.assertIsInstance(field, tag_forms.SingleTagField)
        self.assertEqual(field.tag_options.case_sensitive, True)

        # Check that the option is passed to the widget
        self.assertEqual(field.widget.tag_options.case_sensitive, True)

        # No effect on form data
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["case_sensitive_true"], "mr")

        # No effect on save
        obj = form.save()
        self.assertEqual(obj.case_sensitive_true, "mr")
        self.assertTagModel(tag_model, {"Mr": 0, "mr": 1})

    @skip_if_mysql
    def test_case_sensitive_false(self):
        "Test form SingleTagField case_sensitive false"
        # Prep tag model
        tag_model = self.model.case_sensitive_false.tag_model
        tag_model.objects.get_or_create(name="Mr")

        # Check that the option is passed to the form field from model field
        form = self.form(data={"case_sensitive_false": "mr"})
        field = form.fields["case_sensitive_false"]
        self.assertIsInstance(field, tag_forms.SingleTagField)
        self.assertEqual(field.tag_options.case_sensitive, False)
        self.assertEqual(field.widget.tag_options.case_sensitive, False)

        # Still no effect on form data
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["case_sensitive_false"], "mr")

        # But when saved, the instance will be capitalised
        obj = form.save()
        self.assertEqual(obj.case_sensitive_false, "Mr")
        self.assertTagModel(tag_model, {"Mr": 1})

    def test_force_lowercase_true(self):
        "Test form SingleTagField force_lowercase true"
        # Prep tag model
        tag_model = self.model.force_lowercase_true.tag_model

        # Check that the option is passed to the form field from model field
        form = self.form(data={"force_lowercase_true": "Mr"})
        field = form.fields["force_lowercase_true"]
        self.assertIsInstance(field, tag_forms.SingleTagField)
        self.assertEqual(field.tag_options.force_lowercase, True)
        self.assertEqual(field.widget.tag_options.force_lowercase, True)

        # Form data returns lowercase case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["force_lowercase_true"], "mr")

        # When saved, the instance will be lowercase
        obj = form.save()
        self.assertEqual(obj.force_lowercase_true, "mr")
        self.assertTagModel(tag_model, {"mr": 1})

    def test_force_lowercase_false(self):
        "Test form SingleTagField force_lowercase false"
        # Check that the option is passed to the form field from model field
        form = self.form(data={"force_lowercase_false": "Mr"})
        field = form.fields["force_lowercase_false"]
        self.assertIsInstance(field, tag_forms.SingleTagField)
        self.assertEqual(field.tag_options.force_lowercase, False)
        self.assertEqual(field.widget.tag_options.force_lowercase, False)

        # Check the input string is returned in the new case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["force_lowercase_false"], "Mr")

    def test_max_count(self):
        "Test form SingleTagField max_count is 1"
        # It's always 1. Check any field.
        form = self.form()
        field = form.fields["case_sensitive_true"]
        self.assertIsInstance(field, tag_forms.SingleTagField)
        self.assertEqual(field.tag_options.max_count, 1)
        self.assertEqual(field.widget.tag_options.max_count, 1)

    def test_initial_without_autocomplete_initial(self):
        "Check a field with initial but without autocomplete_initial lists all"
        tag_model = self.model.initial_string.tag_model
        tag_model.objects.create(name="Miss")
        tag_model.objects.create(name="Mx")

        # Confirm default
        field = self.form().fields["initial_string"]
        self.assertSequenceEqual(
            [t.name for t in field.autocomplete_tags],
            [t.name for t in tag_model.objects.all()],
        )

    def test_initial_with_autocomplete_initial(self):
        "Check a field with initial and autocomplete_initial lists initial"
        tag_model = self.model.initial_list.tag_model
        tag_model.objects.create(name="Miss")
        tag_model.objects.create(name="Mx")

        # Confirm default
        field = self.form().fields["initial_list"]
        self.assertSequenceEqual(
            [t.name for t in field.autocomplete_tags],
            [t.name for t in tag_model.objects.initial()],
        )



from django.apps import apps
from django.core.management.base import BaseCommand

from ...models.initial import field_initialise_tags, model_initialise_tags


class Command(BaseCommand):
    """
    Load initial tagulous tags
    Optional argument allows you to specify what to load
    If field_name is missing, initialise all tag fields on the model
    If model_name is missing, initialise all tag fields on models in the app
    If app_name is missing, initialise all tag fields on all models in all apps
    """

    help = "Load initial tagulous tags"

    def add_arguments(self, parser):
        parser.add_argument(
            "--target",
            type=str,
            default="",
            help="Target to load: [<app_name>[.<model_name>[.<field_name>]]]",
        )

    def handle(self, target="", **options):
        # Split up target argument
        parts = target.split(".")
        app_name, model_name, field_name = parts + ([None] * (3 - len(parts)))

        # Look up app
        if app_name:
            app = apps.get_app_config(app_name)
        else:
            app = None

        # Look up specific model, or get all models for the app
        if model_name:
            models = [app.get_model(model_name)]
        elif app is None:
            # Get all models for all apps
            models = apps.get_models()
        else:
            models = app.get_models()

        # If field is specified, can finish here
        if field_name:
            model = models[0]
            field = model._meta.get_field(field_name)
            loaded = field_initialise_tags(model, field, report=self.stdout)

            if not loaded:
                self.stdout.write(
                    "Nothing to load for %s.%s.%s\n"
                    % (model._meta.app_label, model.__name__, field.name)
                )
            return

        # Step through all models and load
        for model in models:
            model_initialise_tags(model, report=self.stdout)



"""
Test models for capitalised app names (#60)
"""
from django.db import models

import tagulous


class CapitalisedTest(models.Model):
    """
    For tests which need a capitalised table name
    """

    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(blank=True)
    tags = tagulous.models.TagField(blank=True)



from django import forms

import tagulous
from tests.tagulous_tests_app import models


# Straight form fields


class SingleTagFieldForm(forms.Form):
    singletag = tagulous.forms.SingleTagField()


class TagFieldForm(forms.Form):
    tags = tagulous.forms.TagField()


# SingleTagField model forms


class SingleTagFieldModelForm(forms.ModelForm):
    class Meta:
        model = models.SingleTagFieldModel
        exclude = ()


class SingleTagFieldOptionalModelForm(forms.ModelForm):
    class Meta:
        model = models.SingleTagFieldOptionalModel
        exclude = ()


class SingleTagFieldRequiredModelForm(forms.ModelForm):
    class Meta:
        model = models.SingleTagFieldRequiredModel
        exclude = ()


class SingleTagFieldOptionsModelForm(forms.ModelForm):
    class Meta:
        model = models.SingleTagFieldOptionsModel
        exclude = ()


# TagField model forms


class TagFieldModelForm(forms.ModelForm):
    class Meta:
        model = models.TagFieldModel
        exclude = ()


class TagFieldOptionalModelForm(forms.ModelForm):
    class Meta:
        model = models.TagFieldOptionalModel
        exclude = ()


class TagFieldRequiredModelForm(forms.ModelForm):
    class Meta:
        model = models.TagFieldRequiredModel
        exclude = ()


class TagFieldOptionsModelForm(forms.ModelForm):
    class Meta:
        model = models.TagFieldOptionsModel
        exclude = ()


# Mixed model forms


class MixedNonTagRefModelForm(forms.ModelForm):
    class Meta:
        model = models.MixedNonTagRefTest
        exclude = ()


# Formsets

SimpleMixedSingleInline = forms.models.inlineformset_factory(
    models.SimpleMixedTest.singletag.tag_model,
    models.SimpleMixedTest,
    formset=tagulous.forms.TaggedInlineFormSet,
    fields=("name",),
)



# -*- coding: utf-8 -*-
from django.db import migrations

import tagulous.models


#
# Basic unittest-esque functions
#


def fail(msg):
    # Fail loudly so the message can be picked up by the calling test
    raise AssertionError(msg)


def assertEqual(a, b):
    if a != b:
        fail("%r != %r" % (a, b))


def assertIsSubclass(cls, supercls):
    if not issubclass(cls, supercls):
        fail("%r is not a subclass of %r: %s" % (cls, supercls, cls.__bases__))


def assertIsInstance(obj, cls, msg=None):
    if not isinstance(obj, cls):
        fail("%r is not an instance of %r" % (obj, cls))


#
# Migration
#


def test_tagulous_in_migrations(apps, schema_editor):
    "Test that Tagulous works in data migrations"
    # Find tagged model
    model = apps.get_model("tagulous_tests_migration", "MigrationTestModel")

    # Check classes have been assigned correctly
    # If so, everything else will work as it should
    assertIsSubclass(model, tagulous.models.TaggedModel)

    assertIsInstance(model.singletag, tagulous.models.SingleTagDescriptor)
    assertIsSubclass(model.singletag.tag_model, tagulous.models.BaseTagModel)

    assertIsInstance(model.tags, tagulous.models.TagDescriptor)
    assertIsSubclass(model.tags.tag_model, tagulous.models.BaseTagTreeModel)


class Migration(migrations.Migration):

    dependencies = [("tagulous_tests_migration", "0003_tree")]

    operations = [
        migrations.RunPython(test_tagulous_in_migrations, lambda *a, **k: None)
    ]



# Generated by Django 2.2.12 on 2020-07-04 05:13

from django.db import migrations, models
import django.db.models.deletion
import tagulous.models.fields
import tagulous.models.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Tagulous_CapitalisedTest_tags',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('slug', models.SlugField()),
                ('count', models.IntegerField(default=0, help_text='Internal counter of how many times this tag is in use')),
                ('protected', models.BooleanField(default=False, help_text='Will not be deleted when the count reaches 0')),
            ],
            options={
                'ordering': ('name',),
                'abstract': False,
                'unique_together': {('slug',)},
            },
            bases=(tagulous.models.models.BaseTagModel, models.Model),
        ),
        migrations.CreateModel(
            name='Tagulous_CapitalisedTest_singletag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('slug', models.SlugField()),
                ('count', models.IntegerField(default=0, help_text='Internal counter of how many times this tag is in use')),
                ('protected', models.BooleanField(default=False, help_text='Will not be deleted when the count reaches 0')),
            ],
            options={
                'ordering': ('name',),
                'abstract': False,
                'unique_together': {('slug',)},
            },
            bases=(tagulous.models.models.BaseTagModel, models.Model),
        ),
        migrations.CreateModel(
            name='CapitalisedTest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('singletag', tagulous.models.fields.SingleTagField(_set_tag_meta=True, blank=True, force_lowercase=True, initial='coding, eating, gaming', null=True, on_delete=django.db.models.deletion.CASCADE, to='tagulousTestsApp3.Tagulous_CapitalisedTest_singletag')),
                ('tags', tagulous.models.fields.TagField(_set_tag_meta=True, blank=True, force_lowercase=True, help_text='Enter a comma-separated tag string', initial='coding, eating, gaming', to='tagulousTestsApp3.Tagulous_CapitalisedTest_tags')),
            ],
        ),
    ]



"""
Tagulous model field managers

These are accessed via the descriptors, and do the work of storing and loading
the tags.

For tag model manager, look in tagulous.models.models
"""
from django.core import exceptions

from ..utils import parse_tags, render_tags


# ##############################################################################
# ###### Manager for SingleTagField
# ##############################################################################


class SingleTagManager(object):
    """
    Manage single tags

    Not a real Django manager; it's a per-instance abstraction between the
    normal FK descriptor to hold in-memory changes of the SingleTagField before
    passing them up to the normal FK descriptor on the pre-save signal.
    """

    def __init__(self, descriptor, instance):
        # The SingleTagDescriptor and instance this manages
        self.descriptor = descriptor
        self.instance = instance

        # Other vars we need
        self.tag_model = self.descriptor.tag_model
        self.field = self.descriptor.field
        self.tag_options = self.descriptor.tag_options

        # Keep track of unsaved changes
        self.changed = False

        # The descriptor stores an unsaved tag string
        # Load the actual value into the cache, if it exists
        # If there is a problem with the actual value, get_actual will fall
        # back to our cache, so make sure it exists first
        self.tag_cache = None
        self.tag_cache = self.get_actual()

        # Start off the local tag name with the actual tag name
        self.tag_name = self.tag_cache.name if self.tag_cache else None

        # Pre/post save will need to keep track of an old tag
        self.removed_tag = None

    def flush_cache(self):
        """
        Clear the FK descriptor's cache

        Must be called after all
        """
        if self.field.is_cached(self.instance):
            self.field.delete_cached_value(self.instance)

    def get_actual(self):
        """
        Get the actual value of the instance according to the FK descriptor
        """
        # A ForeignKey would be on the .attname (field_id), but only
        # if it has been set, otherwise the attribute will not exist.
        #
        # In Django <1.9 we could just check for the attribute, but 1.10
        # sets a DeferredAttribute on the field id to load it on demand, so
        # we need to check the dict directly instead
        check_value = self.field.attname in self.instance.__dict__

        # Check the value if we need to
        if check_value:
            try:
                value = self.descriptor.descriptor.__get__(self.instance)
            except (self.tag_model.DoesNotExist, self.instance.DoesNotExist):
                # Django 1.10 returns instance.DoesNotExist, so check for both.
                #
                # If the tag is deleted but nobody tells this instance; because
                # the real cache is kept empty, this will fail. Try our cache,
                # and if it's set clear the pk.
                self.changed = True
                if self.tag_cache:
                    self.tag_cache.pk = None
                return self.tag_cache

            # The descriptor will have populated the cache, but Django 1.8
            # introduces a check of the cache before allowing save() to start.
            # We can't guarantee that the tag object will have a pk until the
            # pre-save handler has fired - which happens after the cache check.
            # We therefore have to keep the cache clean at all times - we don't
            # lose anything since we keep our own cache, and pre-save will fill
            # it out in time anyway.
            self.flush_cache()
            return value
        return None

    def set_actual(self, value):
        """
        Set the actual value of the instance for the FK descriptor
        """
        self.descriptor.descriptor.__set__(self.instance, value)

        # Django 1.8 cache check fix (see comment in get_actual)
        self.flush_cache()

    def get(self):
        """
        Get the current tag - either a Tag object or None
        If the field has been changed since the instance was last saved, the
        Tag object may be a dynamically generated Tag which does not exist in
        the database yet. The count will not be updated until the instance is
        next saved.
        """
        # If changed, find the tag
        if self.changed:
            if not self.tag_name:
                return None

            # Try to look up the tag
            try:
                if self.tag_options.case_sensitive:
                    tag = self.tag_model.objects.get(name=self.tag_name)
                else:
                    tag = self.tag_model.objects.get(name__iexact=self.tag_name)
            except self.tag_model.DoesNotExist:
                # Does not exist yet, create a temporary one (but don't save)
                if not self.tag_cache:
                    self.tag_cache = self.tag_model(name=self.tag_name, protected=False)
                tag = self.tag_cache
            return tag
        else:
            # Return the response that it should have had (a Tag or None)
            return self.get_actual()

    def set(self, value):
        """
        Set the current tag
        """
        if not value:
            tag_name = ""

        elif isinstance(value, str):
            # Force tag to lowercase
            if self.tag_options.force_lowercase:
                value = value.lower()
            tag_name = value

        # Look up the tag name
        else:
            tag_name = value.name

        # If no change, do nothing
        if self.tag_name == tag_name:
            return

        # Store the tag name and mark changed
        self.changed = True
        self.tag_name = tag_name
        self.tag_cache = None

    def pre_save_handler(self):
        """
        When the model is about to save, update the tag value
        """
        # Get the new tag
        new_tag = self.get()

        # Logic check to replace standard null/blank model field validation
        if not new_tag and self.field.required:
            raise exceptions.ValidationError(self.field.error_messages["null"])

        # Only need to go further if there has been a change
        if not self.changed:
            return

        # Store the old tag so we know to decrement it in post_save
        self.removed_tag = self.get_actual()

        # Create or increment the tag object
        if new_tag:
            # Ensure it is in the database
            if not new_tag.pk:
                new_tag.save()

            # Increment the new tag
            new_tag.increment()

        # Set it
        self.set_actual(new_tag)

        # Clear up
        self.changed = False

    def post_save_handler(self):
        """
        When the model has saved, decrement the old tag
        """
        if self.removed_tag:
            self.removed_tag.decrement()

    def post_delete_handler(self):
        """
        When the model has been deleted, decrement the actual tag
        """
        # Decrement the actual tag
        old_tag = self.get_actual()
        if not old_tag:
            return

        # Try to update the old tag
        try:
            old_tag.decrement()
        except type(old_tag).DoesNotExist:
            # The tag was just deleted along with the model in the same operation - most
            # likely a cascade delete originated on the tag model
            pass

        # Clear the tag on this instance so we don't leave a reference
        self.set_actual(None)

        # If there is no new value, mark the old one as a new one,
        # so the database will be updated if the instance is saved again
        if not self.changed:
            self.tag_name = old_tag.name
        self.tag_cache = None
        self.changed = True


# ##############################################################################
# ###### Mixin for TagField manager
# ##############################################################################


class BaseTagRelatedManager(object):
    """
    Mixin for TagRelatedManagerMixin, and base class for FakeTagRelatedManager.

    Provides methods to managed cached tags
    """

    @property
    def tags(self):
        if not hasattr(self, "_tags") or self._tags is None:
            try:
                self._tags = list(self.all())
            except AttributeError:
                self._tags = []

        return self._tags

    @tags.setter
    def tags(self, value):
        self._tags = value

    def init_tagulous(self, descriptor):
        """
        Called directly after the mixin is added to the instantiated manager
        """
        self.tag_model = descriptor.tag_model
        self.tag_options = descriptor.tag_options

        # Maintain an internal set of tags, and track whether they've changed
        # If internal tags are None, haven't been loaded yet
        self.changed = False
        self.tags = None

    def __str__(self):
        """
        If called on an instance, return the tag string
        """
        return self.get_tag_string()

    def __contains__(self, item):
        item_str = str(item)
        if self.tag_options.force_lowercase:
            item_str = item_str.lower()

        if self.tag_options.case_sensitive:
            return item_str in [tag.name for tag in self.tags]
        return item_str in [tag.name.lower() for tag in self.tags]

    def __eq__(self, other):
        """
        Compare a tag string or iterable of tags to the tags on this manager
        """
        # If not case sensitive, or lowercase forced, compare on lowercase
        lower = False
        if self.tag_options.force_lowercase or not self.tag_options.case_sensitive:
            lower = True

        # Prep other argument we're comparing against
        if isinstance(other, BaseTagRelatedManager):
            other = other.tags
        if isinstance(other, str):
            other_str = str(other)

            # Enforce case non-sensitivity or lowercase
            if lower:
                other_str = other_str.lower()

            # Parse other_str into list of tags
            other_tags = parse_tags(
                other_str, space_delimiter=self.tag_options.space_delimiter
            )

        else:
            # Assume it's an iterable
            other_tags = other
            if lower:
                other_tags = [str(tag).lower() for tag in other]

        # Get list of set tags
        self_tags = self.get_tag_list()

        # Compare tag count
        if len(other_tags) != len(self_tags):
            return False

        # Compare tags
        for tag in self_tags:
            # If lowercase or not case sensitive, lower for comparison
            if lower:
                tag = tag.lower()

            # Check tag in other tags
            if tag not in other_tags:
                return False

        # Same number of tags, and all self tags present in other tags
        # It's a match
        return True

    def __ne__(self, other):
        """
        Compare tags, using opposite of __eq__
        """
        return not self.__eq__(other)

    def load_from_tagmanager(self, manager):
        """
        Copy status and cache from the specified manager
        """
        self.changed = manager.changed
        self.tags = manager.tags

    #
    # Functions for getting and setting tag cache
    #

    def get_tag_string(self):
        """
        Get the tag edit string for this instance as a string
        """
        return render_tags(self.tags)

    def get_tag_list(self):
        """
        Get the tag names for this instance as a list of tag names
        """
        return [tag.name for tag in self.tags]

    def set_tag_string(self, tag_string):
        """
        Sets the tags for this instance, given a tag edit string
        """
        # Get all tag names
        tag_names = parse_tags(
            tag_string, space_delimiter=self.tag_options.space_delimiter
        )

        # Pass on to set_tag_list
        return self.set_tag_list(tag_names)

    set_tag_string.alters_data = True

    def set_tag_list(self, tag_names):
        """
        Sets the tags for this instance, given a list of tag names, or a list
        or queryset of tags
        """
        if self.tag_options.max_count and len(tag_names) > self.tag_options.max_count:
            raise ValueError(
                "Cannot set more than %d tags on this field"
                % self.tag_options.max_count
            )

        # Force tag_names to strings, in case it's a list of tags or a queryset
        tag_names = [str(tag_name) for tag_name in tag_names]

        # Apply force_lowercase
        if self.tag_options.force_lowercase:
            # Will be lowercase for later comparison
            tag_names = [name.lower() for name in tag_names]

        # Prep tag lookup
        # old_tags      = { cmp_name: tag }
        # cmp_new_names = { cmp_name: cased_name }
        if self.tag_options.case_sensitive:
            old_tags = dict([(tag.name, tag) for tag in self.tags])
            cmp_new_names = dict([(n, n) for n in tag_names])
        else:
            # Not case sensitive - need to compare on lowercase
            old_tags = dict([(tag.name.lower(), tag) for tag in self.tags])
            cmp_new_names = dict([(name.lower(), name) for name in tag_names])

        # See which tags are staying
        new_tags = []
        for cmp_old_name, old_tag in old_tags.items():
            if cmp_old_name in cmp_new_names:
                # Exists - add to new tags
                new_tags.append(old_tag)
                del cmp_new_names[cmp_old_name]
            else:
                # Tag will be removed
                self.changed = True

        # Only left with tag names which aren't present
        for tag_name in cmp_new_names.values():
            # Find or create all new tags
            try:
                if self.tag_options.case_sensitive:
                    tag = self.tag_model.objects.get(name=tag_name)
                else:
                    tag = self.tag_model.objects.get(name__iexact=tag_name)
            except self.tag_model.DoesNotExist:
                # Don't create it until it's saved
                tag = self.tag_model(name=tag_name, protected=False)

            # Add the tag
            new_tags.append(tag)
            self.changed = True

        # Store in internal tag cache
        self.tags = new_tags

    set_tag_list.alters_data = True


class FakeTagRelatedManager(BaseTagRelatedManager):
    """
    Fake manager class to manage cached tags, but provide no database access.

    For use with an unsaved model instance
    """

    _needs_db = '"%r" needs to be saved before TagField can use the database'

    def __init__(self, descriptor, instance, instance_type):
        self.instance = instance
        self.model = instance.__class__
        self.init_tagulous(descriptor)

    def reload(self):
        """
        Instance does not exist in the database, so should wipe the tags
        """
        self.tags = []
        self.changed = False

    def save(self, force=False):
        raise ValueError(self._needs_db % self.instance)

    def set(self, *args):
        raise ValueError(self._needs_db % self.instance)

    def add(self, *args):
        raise ValueError(self._needs_db % self.instance)

    def remove(self, *args):
        raise ValueError(self._needs_db % self.instance)

    def clear(self):
        raise ValueError(self._needs_db % self.instance)


class TagRelatedManagerMixin(BaseTagRelatedManager):
    """
    Mixin for RelatedManager to add tag functions

    Added to the normal m2m RelatedManager, after it has been instantiated.
    This holds in-memory changes of the TagField before committing them to the
    database on the post-save signal.
    """

    def reload(self):
        """
        Get the actual tags
        """
        # Convert to a list to force it to load now, and so we can change it
        self.tags = list(self.all())
        self.changed = False

    reload.alters_data = True

    def post_save_handler(self):
        """
        When the model has saved, save related tags

        Called by the signal handler
        """
        self.save()

    def pre_delete_handler(self):
        """
        Safely clear the M2M tag field, persisting tags on the manager

        When deleting an instance, Django does not call the add/remove of the
        M2M manager, so tag counts would be too high. Related to:
          https://code.djangoproject.com/ticket/6707
        We need to clear the M2M manually to update tag counts

        Called by the signal handler
        """
        # Get tags so we can make them available to the fake manager later
        self.reload()
        tags = self.tags

        # Clear the object
        self.clear()

        # Put the tags back on the manager
        self.tags = tags

    def save(self, force=False):
        """
        Set the actual tags to the internal tag state

        If force is True, save whether we think it has changed or not
        """
        if not self.changed and not force:
            return

        # Add and remove tags as necessary
        new_tags = self._ensure_tags_in_db(self.tags)
        self.reload()

        # Add new tags
        for new_tag in new_tags:
            if new_tag not in self.tags:
                self.add(new_tag, _enforce_max_count=False)

        # Remove old tags
        for old_tag in self.tags:
            if old_tag not in new_tags:
                self.remove(old_tag)
        self.tags = new_tags
        self.changed = False

    save.alters_data = True

    def _ensure_tags_in_db(self, tags):
        """
        Ensure that self.tags all exist in the database
        """
        db_tags = []
        for tag in tags:
            if tag.pk:
                # Already in DB
                db_tag = tag
            else:
                # Not in DB - get or create
                field_lookup = "name"
                if not self.tag_options.case_sensitive:
                    field_lookup += "__iexact"
                db_tag, __ = self.tag_model.objects.get_or_create(
                    defaults={"name": tag.name, "protected": False},
                    **{field_lookup: tag.name}
                )
            db_tags.append(db_tag)
        return db_tags

    #
    # New set, add, remove and clear, to update tag counts
    #
    def set(self, *objs):
        self.clear()
        self.add(*objs)

    def add(self, *objs, **kwargs):
        """
        Add a list of tags or tag strings

        Takes on internal argument, _enforce_max_count - don't use in your code
        """
        enforce_max_count = kwargs.pop("_enforce_max_count", True)
        if kwargs:
            raise TypeError("add() got an unexpected keyword argument")

        # Convert strings to tag objects
        new_tags = []
        for tag in objs:
            if isinstance(tag, str):
                new_tags.append(self.tag_model(name=tag))
            else:
                new_tags.append(tag)

        # Don't trust the internal tag cache
        self.reload()

        # Reduce tags to ones not already loaded
        new_tags = [tag for tag in new_tags if tag not in self.tags]

        # Enforce max_count
        if enforce_max_count and self.tag_options.max_count:
            current_count = len(self.tags)
            if current_count + len(new_tags) > self.tag_options.max_count:
                raise ValueError(
                    "Cannot set more than %s tags on this field; it already has %s"
                    % (self.tag_options.max_count, current_count)
                )

        # Ensure tags exist
        new_tags = self._ensure_tags_in_db(new_tags)

        # Add to db, add to cache, and increment
        super(TagRelatedManagerMixin, self).add(*new_tags)
        for tag in new_tags:
            self.tags.append(tag)
            tag.increment()

    add.alters_data = True

    def remove(self, *objs):
        # Convert strings to tag objects - if object doesn't exist, skip
        rm_tags = []
        for tag in objs:
            if isinstance(tag, str):
                try:
                    rm_tags.append(self.tag_model.objects.get(name=tag))
                except self.tag_model.DoesNotExist:
                    continue
            else:
                rm_tags.append(tag)

        # Don't trust the internal tag cache
        self.reload()

        # Cut tags back to only ones already set
        rm_tags = [tag for tag in self.tags if tag in rm_tags]

        # Remove from cache
        self.tags = [tag for tag in self.tags if tag not in rm_tags]

        # Remove from db and decrement
        super(TagRelatedManagerMixin, self).remove(*self._ensure_tags_in_db(rm_tags))
        for tag in rm_tags:
            tag.decrement()

    remove.alters_data = True

    def clear(self):
        # Don't trust the internal tag cache
        self.reload()

        # Clear db, then decrement and empty cache
        super(TagRelatedManagerMixin, self).clear()
        for tag in self.tags:
            tag.decrement()
        self.tags = []

    clear.alters_data = True



"""
Model signal handlers

These are connected in tagulous.apps.TagulousConfig.ready()
"""

from ..models.fields import SingleTagField, TagField
from ..models.tagged import TaggedModel


class TaggedSignalHandler(object):
    """
    Base class handling signals from TaggedModel subclasses
    """

    def __call__(self, sender, instance, **kwargs):
        if not self.is_relevant(sender):
            return

        for field, field_type in self.get_fields(sender):
            descriptor = getattr(sender, field.name)
            manager = descriptor.get_manager(instance)

            self.handle(manager, field_type, kwargs.get("raw", False))

    def is_relevant(self, sender):
        return issubclass(sender, TaggedModel)

    def get_fields(self, sender):
        """
        Generator which returns (field, field_type) pairs
        """
        for field in sender._meta.get_fields():
            if isinstance(field, SingleTagField):
                yield (field, SingleTagField)
            elif isinstance(field, TagField):
                yield (field, TagField)

    def handle(self, manager, field_type, is_raw):
        raise NotImplementedError()  # noqa


class PreSaveHandler(TaggedSignalHandler):
    """
    Pre-save signal handler

    Make sure SingleTagField is in correct format to be saved. The manager needs to know
    when the model is about to be saved # so that it can ensure the tag exists and
    assign its pk to field_id.
    """

    def handle(self, manager, field_type, is_raw):
        if field_type != SingleTagField:
            return
        manager.pre_save_handler()


class PostSaveHandler(TaggedSignalHandler):
    """
    Post-save signal handler

    Ensure both tag fields' states are saved
    """

    def handle(self, manager, field_type, is_raw):
        manager.post_save_handler()

        # If raw is set, data is being injected into the system, most likely from a
        # deserialization operation. If the tag model has just been deserialized too,
        # the tag counts will probably be off.
        if is_raw:
            if field_type == SingleTagField:
                manager.get().update_count()
            else:
                for tag in manager.tags:
                    tag.update_count()


class PropagatedSignalMixin(object):
    def get_fields(self, sender):
        """
        Post delete signal propagates up the class hierarchy, so filter fields
        returned to those belonging to the model which raised this signal
        """
        for field, field_type in super(PropagatedSignalMixin, self).get_fields(sender):
            if field.model != sender:
                continue
            yield (field, field_type)


class PreDeleteHandler(PropagatedSignalMixin, TaggedSignalHandler):
    """
    Pre-delete signal handler

    Ensure TagField is cleaned out before the main object is deleted
    """

    def handle(self, manager, field_type, is_raw):
        if field_type != TagField:
            return
        manager.pre_delete_handler()


class PostDeleteHandler(PropagatedSignalMixin, TaggedSignalHandler):
    """
    Post-delete signal handler

    Update tag count on delete
    """

    def handle(self, manager, field_type, is_raw):
        if field_type != SingleTagField:
            return
        manager.post_delete_handler()


def register_post_signals():
    from django.db.models.signals import post_delete, post_save, pre_delete, pre_save

    pre_save.connect(PreSaveHandler(), weak=False)
    post_save.connect(PostSaveHandler(), weak=False)
    pre_delete.connect(PreDeleteHandler(), weak=False)
    post_delete.connect(PostDeleteHandler(), weak=False)
